#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import pytest

from paleomix.common.formats.newick import (
    GraphError,
    Newick,
    NewickError,
    NewickParseError,
)

###############################################################################
###############################################################################
# Constructor


def test_newick__constructor__name():
    node = Newick(name="AbC")
    assert node.name == "AbC"


def test_newick__constructor__children_set_in_internal_nodes():
    node = Newick(name="Leaf")
    top_node = Newick(children=[node])
    assert top_node.children == (node,)


def test_newick__constructor__children_not_set_in_leaf_nodes():
    node = Newick(name="Leaf")
    assert node.children == ()


def test_newick__constructor__is_leaf_true_for_leaf_nodes():
    node = Newick(name="Another Leaf")
    assert node.is_leaf


def test_newick__constructor__is_leaf_false_for_internal_nodes():
    node = Newick(name="Leaf")
    top_node = Newick(children=[node])
    assert not top_node.is_leaf


def test_newick__constuctor__leaf_nodes_must_have_name_or_length():
    with pytest.raises(NewickError):
        Newick(children=None)


def test_newick__constructor__internal_nodes_must_have_children():
    with pytest.raises(NewickError):
        Newick(children=[])


def test_newick__constructor__children_must_be_newick():
    with pytest.raises(TypeError):
        Newick(children=["A", "B"])


###############################################################################
###############################################################################
# get_leaf_nodes


def test_newick__get_leaf_nodes__leaf_returns_self():
    node = Newick(name="Leaf")
    assert list(node.get_leaf_nodes()) == list([node])


def test_newick__get_leaf_nodes__internal_node_returns_leaf_nodes():
    node_a = Newick(name="Leaf A")
    node_b = Newick(name="Leaf B")
    top_node = Newick(children=[node_a, node_b])
    assert list(top_node.get_leaf_nodes()) == list([node_a, node_b])


def test_newick__get_leaf_nodes__complex_case():
    node_a = Newick(name="Leaf A")
    node_b = Newick(name="Leaf B")
    node_c = Newick(name="Leaf C")
    sub_a = Newick(children=[node_b, node_c])
    top_node = Newick(children=[node_a, sub_a])
    assert list(top_node.get_leaf_nodes()) == list([node_a, node_b, node_c])


###############################################################################
###############################################################################
# get_leaf_nodes


def test_newick__get_leaf_names__leaf_returns_self():
    node = Newick(name="Leaf")
    assert list(node.get_leaf_names()) == list(["Leaf"])


def test_newick__get_leaf_names__internal_node_returns_leaf_nodes():
    node_a = Newick(name="Leaf A")
    node_b = Newick(name="Leaf B")
    top_node = Newick(children=[node_a, node_b])
    assert list(top_node.get_leaf_names()) == list(["Leaf A", "Leaf B"])


def test_newick__get_leaf_names__complex_case():
    node_a = Newick(name="Leaf A")
    node_b = Newick(name="Leaf B")
    node_c = Newick(name="Leaf C")
    sub_a = Newick(children=[node_b, node_c])
    top_node = Newick(children=[node_a, sub_a])
    assert list(top_node.get_leaf_names()) == list(["Leaf A", "Leaf B", "Leaf C"])


###############################################################################
###############################################################################
# reroot_on_taxa


def test_newick__reroot_on_taxa__single_taxa():
    source = Newick.from_string("((A,B),C);")
    expected = Newick.from_string("((B,C),A);")
    assert expected == source.reroot_on_taxa("A")


def test_newick__reroot_on_taxa__single_taxa_with_branch_lengths():
    source = Newick.from_string("((A:4,B:3):2,C:1);")
    expected = Newick.from_string("((B:3,C:3.0):2.0,A:2.0);")
    assert expected == source.reroot_on_taxa("A")


def test_newick__reroot_on_taxa__multiple_taxa__clade():
    source = Newick.from_string("((A,(B,C)),(D,E));")
    expected = Newick.from_string("(((D,E),A),(B,C));")
    assert expected == source.reroot_on_taxa(("B", "C"))


def test_newick__reroot_on_taxa__multiple_taxa__paraphylogeny():
    source = Newick.from_string("((B,C),((D,E),A));")
    expected = Newick.from_string("(((B,C),A),(D,E));")
    assert expected == source.reroot_on_taxa(("A", "C"))


def test_newick__reroot_on_taxa__no_taxa():
    source = Newick.from_string("((B,C),((D,E),A));")
    with pytest.raises(ValueError):
        source.reroot_on_taxa(())


def test_newick__reroot_on_taxa__unknown_taxa():
    source = Newick.from_string("((B,C),((D,E),A));")
    with pytest.raises(ValueError):
        source.reroot_on_taxa(("A", "Z"))


def test_newick__reroot_on_taxa__no_non_root_taxa():
    source = Newick.from_string("((B,C),((D,E),A));")
    with pytest.raises(ValueError):
        source.reroot_on_taxa(("A", "B", "C", "D", "E"))


###############################################################################
###############################################################################
# reroot_on_midpoint


def test_newick__reroot_on_midpoint__single_node():
    source = Newick.from_string("(A:3.0);")
    expected = Newick.from_string("(A:3.0);")
    assert expected == source.reroot_on_midpoint()


def test_newick__reroot_on_midpoint__two_nodes():
    source = Newick.from_string("(A:3.0,B:8.0);")
    rerooted = source.reroot_on_midpoint()
    expected = Newick.from_string("(A:5.5,B:5.5);")
    assert expected == rerooted


def test_newick__reroot_on_midpoint__two_clades():
    source = Newick.from_string("((A:7,B:2):1,(C:1,D:0.5):2);")
    rerooted = source.reroot_on_midpoint()
    expected = Newick.from_string("(((C:1,D:0.5):3.0,B:2):1.5,A:5.5);")
    assert expected == rerooted


def test_newick__reroot_on_midpoint__nested_clades():
    source = Newick.from_string("((A:2,(B:2,C:3):4):1,(D:1,E:0.5):2);")
    rerooted = source.reroot_on_midpoint()
    expected = Newick.from_string("(((D:1,E:0.5):3.0,A:2):1.5,(B:2,C:3):2.5);")
    assert expected == rerooted


def test_newick__reroot_on_midpoint__reroot_on_internal_node():
    source = Newick.from_string("((A:5.0,B:1.0)C:2.0,D:3.0);")
    rerooted = source.reroot_on_midpoint()
    expected = Newick.from_string("(A:5.0,B:1.0,D:5.0)C;")
    assert expected == rerooted


_INVALID_BRANCH_LENGTHS = (
    "(A,B);",  # No branch lengths
    "(A:7,B);",  # Length missing for leaf node
    "(A:7,(B:3));",  # Length missing for inner node
    "(A:7,(B:3):-1);",  # Negative branch length
    "(A:7,B:-1);",  # Negative leaf length
)


@pytest.mark.parametrize("newick", _INVALID_BRANCH_LENGTHS)
def test_newick__reroot_on_midpoint__invalid_branch_lengths(newick):
    source = Newick.from_string(newick)
    with pytest.raises(GraphError):
        source.reroot_on_midpoint()


###############################################################################
###############################################################################
# add_support


def test_newick__add_support__no_trees():
    main_tree = Newick.from_string("(((A,B),C),D);")
    expected = Newick.from_string("(((A,B)0,C)0,D);")
    result = main_tree.add_support([])
    assert expected == result


def test_newick__add_support__single_identical_tree():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [Newick.from_string("(((A,B),C),D);")]
    expected = Newick.from_string("(((A,B)1,C)1,D);")
    result = main_tree.add_support(bootstraps)
    assert expected == result


def test_newick__add_support__single_identical_tree__different_rooting():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [Newick.from_string("(((C,D),B),A);")]
    expected = Newick.from_string("(((A,B)1,C)1,D);")
    result = main_tree.add_support(bootstraps)
    assert expected == result


def test_newick__add_support__multiple_trees__different_topologies():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [
        Newick.from_string("(((C,B),D),A);"),
        Newick.from_string("(((A,D),B),C);"),
    ]
    expected = Newick.from_string("(((A,B)0,C)2,D);")
    result = main_tree.add_support(bootstraps)
    assert expected == result


def test_newick__add_support__multiple_trees__partially_different_topologies():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [
        Newick.from_string("(((C,D),A),B);"),
        Newick.from_string("(((A,D),B),C);"),
    ]
    expected = Newick.from_string("(((A,B)1,C)2,D);")
    result = main_tree.add_support(bootstraps)
    assert expected == result


def test_newick__add_support__multiple_trees__two_cladees():
    main_tree = Newick.from_string("((A,B),(C,(D,E)));")
    bootstraps = [
        Newick.from_string("((((C,E),D),A),B);"),
        Newick.from_string("(((A,(C,D)),B),E);"),
    ]
    expected = Newick.from_string("((A,B)1,(C,(D,E)0)1);")
    result = main_tree.add_support(bootstraps)
    assert expected == result


def test_newick__add_support__differing_leaf_names():
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [Newick.from_string("(((C,E),B),A);")]
    with pytest.raises(NewickError):
        main_tree.add_support(bootstraps)


_ADD_SUPPORT_FORMATTING = (
    ("{Support}", "(((A,B)1,C)3,D);"),
    ("{Percentage:.0f}", "(((A,B)33,C)100,D);"),
    ("{Fraction:.2f}", "(((A,B)0.33,C)1.00,D);"),
)


@pytest.mark.parametrize("fmt, expected", _ADD_SUPPORT_FORMATTING)
def test_newick__add_support__formatting(fmt, expected):
    main_tree = Newick.from_string("(((A,B),C),D);")
    bootstraps = [
        Newick.from_string("(((C,D),A),B);"),
        Newick.from_string("(((C,B),A),D);"),
        Newick.from_string("(((A,D),B),C);"),
    ]
    expected = Newick.from_string(expected)
    result = main_tree.add_support(bootstraps, fmt)
    assert expected == result


def test_newick__add_support__unique_names_required():
    main_tree = Newick.from_string("(((A,B),C),A);")
    bootstraps = [Newick.from_string("(((A,B),C),A);")]
    with pytest.raises(NewickError):
        main_tree.add_support(bootstraps)


###############################################################################
###############################################################################
# from_string


def test_newick__parse__minimal_newick__name_only():
    top_node = Newick(name="A")
    assert Newick.from_string("A;") == top_node


def test_newick__parse__single_taxa():
    child_node = Newick(name="Ab")
    top_node = Newick(children=[child_node])
    assert Newick.from_string("(Ab);") == top_node


def test_newick__parse__two_taxa():
    child_node_1 = Newick(name="A")
    child_node_2 = Newick(name="Bc")
    top_node = Newick(children=[child_node_1, child_node_2])
    assert Newick.from_string("(A,Bc);") == top_node


def test_newick__parse__three_taxa():
    child_node_1 = Newick(name="A")
    child_node_2 = Newick(name="Bc")
    child_node_3 = Newick(name="DeF")
    top_node = Newick(children=[child_node_1, child_node_2, child_node_3])
    assert Newick.from_string("(A,Bc,DeF);") == top_node


def test_newick__parse__ignore_whitespace():
    assert Newick.from_string("(A,B);") == Newick.from_string("(A, B);")


def test_newick__parse__missing_semicolon():
    with pytest.raises(NewickParseError):
        Newick.from_string("()")


def test_newick__parse__subnode__single_taxa():
    child_node_1 = Newick(name="A")
    child_node_2a = Newick(name="B")
    child_node_2 = Newick(children=[child_node_2a])
    top_node = Newick(children=[child_node_1, child_node_2])
    assert Newick.from_string("(A,(B));") == top_node


def test_newick__parse__subnode__two_taxa():
    child_node_1 = Newick(name="A")
    child_node_2a = Newick(name="B")
    child_node_2b = Newick(name="C")
    child_node_2 = Newick(children=[child_node_2a, child_node_2b])
    top_node = Newick(children=[child_node_1, child_node_2])
    assert Newick.from_string("(A,(B,C));") == top_node


###########################################################################
###########################################################################
# cmp - white-box, just make sure all properties are compared


def test_newick__cmp__identical():
    node_a = Newick(name="A", length=13, children=[Newick(name="B")])
    node_b = Newick(name="A", length=13, children=[Newick(name="B")])
    assert node_a == node_b


def test_newick__cmp__identical_for_empty_string_length():
    node_a = Newick(name="A", length="", children=[Newick(name="B")])
    node_b = Newick(name="A", length=None, children=[Newick(name="B")])
    assert node_a == node_b


def test_newick__cmp__identical_for_empty_string_name():
    node_a = Newick(name="", length=13, children=[Newick(name="B")])
    node_b = Newick(name=None, length=13, children=[Newick(name="B")])
    assert node_a == node_b


_CMP_NOT_IDENTICAL = (
    Newick(name="B", length=13, children=[Newick(name="B")]),
    Newick(name="A", length=14, children=[Newick(name="B")]),
    Newick(name="A", length=13, children=[]),
    Newick(name="A", length=13, children=[Newick(name="C")]),
    Newick(name="B", length=14, children=[Newick(name="C")]),
)


@pytest.mark.parametrize("node_b", _CMP_NOT_IDENTICAL)
def test_newick__cmp__not_identical(node_b):
    node_a = Newick(name="A", length=13, children=[Newick(name="B")])
    assert node_a != node_b


###############################################################################
###############################################################################
# hash - white-box, just make sure all properties are used


def test_newick__hash__identical():
    node_a = Newick(name="A", length=13, children=[Newick(name="B")])
    node_b = Newick(name="A", length=13, children=[Newick(name="B")])
    assert hash(node_a) == hash(node_b)


_HASH_NOT_IDENTICAL = (
    Newick(name="B", length=13, children=[Newick(name="B")]),
    Newick(name="A", length=14, children=[Newick(name="B")]),
    Newick(name="A", length=13, children=[]),
    Newick(name="A", length=13, children=[Newick(name="C")]),
    Newick(name="B", length=14, children=[Newick(name="C")]),
)


@pytest.mark.parametrize("node_b", _CMP_NOT_IDENTICAL)
def test_newick__hash__not_identical(node_b):
    node_a = Newick(name="A", length=13, children=[Newick(name="B")])
    assert hash(node_a) != hash(node_b)


def test_newick__hash__hashable():
    key_a = Newick(name="A", length=13.7, children=[Newick(name="F")])
    key_b = Newick(name="A", length=13.7, children=[Newick(name="F")])
    assert key_b in {key_a: True}


###############################################################################
###############################################################################
# Malformed newick strings


def test_newick__malformed__unbalanced_parantheses():
    with pytest.raises(NewickParseError):
        Newick.from_string("(A,(B,C);")


def test_newick__malformed__mismatched_parantheses():
    with pytest.raises(NewickParseError):
        Newick.from_string("(A,(B,C();")


def test_newick__malformed__missing_parantheses():
    with pytest.raises(NewickParseError):
        Newick.from_string("(A,(B,C))")


def test_newick__malformed__missing_length():
    with pytest.raises(NewickParseError):
        Newick.from_string("(A:,(B,C));")
    with pytest.raises(NewickParseError):
        Newick.from_string("(A,(B:,C));")
    with pytest.raises(NewickParseError):
        Newick.from_string("(A,(B,C:));")
    with pytest.raises(NewickParseError):
        Newick.from_string("(A,(B,C):);")
    with pytest.raises(NewickParseError):
        Newick.from_string("(A,(B,C)):;")


def test_newick__malformed__multiple_lengths():
    with pytest.raises(NewickParseError):
        Newick.from_string("(A:1:2);")


###############################################################################
###############################################################################
# Implicit leafs are not supported (due to problems with ambiguiety)


def test_newick__parse__first_taxa_unnamed():
    with pytest.raises(NewickError):
        Newick.from_string("(,A);")


def test_newick__parse__second_taxa_unnamed():
    with pytest.raises(NewickError):
        Newick.from_string("(A,);")


def test_newick__parse__two_taxa_unnamed():
    with pytest.raises(NewickError):
        Newick.from_string("(,);")


def test_newick__parse__three_taxa_unnamed():
    with pytest.raises(NewickError):
        Newick.from_string("(,,);")


###############################################################################
###############################################################################
# Empty non-leaf nodes are not allowed, as their interpretation is unclear


def test_newick__parse__minimal_newick__implicit_nodes():
    with pytest.raises(NewickParseError):
        Newick.from_string("();")


def test_newick__parse__subnode__empty():
    with pytest.raises(NewickParseError):
        Newick.from_string("(A,());")


###############################################################################
###############################################################################
# The following tests are derived from the wikipedia description of the
# newick format: http://en.wikipedia.org/wiki/Newick_format#Examples


def test_newick__wikipedia_example_1():
    # no nodes are named, this format is not supported here!
    with pytest.raises(NewickError):
        Newick.from_string("(,,(,));")


def test_newick__wikipedia_example_2():
    # leaf nodes are named
    taxa_d = Newick(name="D")
    taxa_c = Newick(name="C")
    taxa_sub = Newick(children=[taxa_c, taxa_d])
    taxa_b = Newick(name="B")
    taxa_a = Newick(name="A")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub])
    assert Newick.from_string("(A,B,(C,D));") == top_node


def test_newick__wikipedia_example_3():
    # all nodes are named
    taxa_d = Newick(name="D")
    taxa_c = Newick(name="C")
    taxa_sub = Newick(children=[taxa_c, taxa_d], name="E")
    taxa_b = Newick(name="B")
    taxa_a = Newick(name="A")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub], name="F")
    assert Newick.from_string("(A,B,(C,D)E)F;") == top_node


def test_newick__wikipedia_example_4():
    # all but root node have a distance to parent
    taxa_d = Newick(length="0.4")
    taxa_c = Newick(length="0.3")
    taxa_sub = Newick(children=[taxa_c, taxa_d], length="0.5")
    taxa_b = Newick(length="0.2")
    taxa_a = Newick(length="0.1")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub])
    assert Newick.from_string("(:0.1,:0.2,(:0.3,:0.4):0.5);") == top_node


def test_newick__wikipedia_example_5():
    # all have a distance to parent
    taxa_d = Newick(length="0.4")
    taxa_c = Newick(length="0.3")
    taxa_sub = Newick(children=[taxa_c, taxa_d], length="0.5")
    taxa_b = Newick(length="0.2")
    taxa_a = Newick(length="0.1")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub], length="0.0")
    assert Newick.from_string("(:0.1,:0.2,(:0.3,:0.4):0.5):0.0;") == top_node


def test_newick__wikipedia_example_6():
    # distances and leaf names (popular)
    taxa_d = Newick(length="0.4", name="D")
    taxa_c = Newick(length="0.3", name="C")
    taxa_sub = Newick(children=[taxa_c, taxa_d], length="0.5")
    taxa_b = Newick(length="0.2", name="B")
    taxa_a = Newick(length="0.1", name="A")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub])
    assert Newick.from_string("(A:0.1,B:0.2,(C:0.3,D:0.4):0.5);") == top_node


def test_newick__wikipedia_example_7():
    # distances and all names
    taxa_d = Newick(length="0.4", name="D")
    taxa_c = Newick(length="0.3", name="C")
    taxa_sub = Newick(children=[taxa_c, taxa_d], length="0.5", name="E")
    taxa_b = Newick(length="0.2", name="B")
    taxa_a = Newick(length="0.1", name="A")
    top_node = Newick(children=[taxa_a, taxa_b, taxa_sub], name="F")
    assert Newick.from_string("(A:0.1,B:0.2,(C:0.3,D:0.4)E:0.5)F;") == top_node


def test_newick__wikipedia_example_8():
    # a tree rooted on a leaf node (rare)
    taxa_b = Newick(length="0.2", name="B")
    taxa_c = Newick(length="0.3", name="C")
    taxa_d = Newick(length="0.4", name="D")
    node_e = Newick(length="0.5", name="E", children=[taxa_c, taxa_d])
    node_f = Newick(length="0.1", name="F", children=[taxa_b, node_e])
    node_a = Newick(name="A", children=[node_f])
    assert Newick.from_string("((B:0.2,(C:0.3,D:0.4)E:0.5)F:0.1)A;") == node_a


###############################################################################
###############################################################################
# str / repr


def test_newick__str__non_string_name():
    node = Newick(children=[Newick(name=17, length="1.3")])
    assert str(node) == "(17:1.3);"


def test_newick__str__non_string_length():
    node = Newick(children=[Newick(name="Foo", length=1.3)])
    assert str(node) == "(Foo:1.3);"


def test_newick__str__repr_equal_to_str():
    node_a = Newick(name="A", length="123")
    node_b = Newick(name="B", length="4567")
    top_node = Newick(children=[node_a, node_b])
    assert str(top_node) == "(A:123,B:4567);"


def test_newick__str__single_leaf_should_not_be_followed_by_comma():
    node = Newick(name="A")
    top_node = Newick(children=[node])
    assert str(top_node) == "(A);"


_STR_EQUALITY = (
    # 2. leaf nodes are named
    "(A,B,(C,D));",
    # 3. all nodes are named
    "(A,B,(C,D)E)F;",
    # 4. all but root node have a distance to parent
    "(:0.1,:0.2,(:0.3,:0.4):0.5);",
    # 5. all have a distance to parent
    "(:0.1,:0.2,(:0.3,:0.4):0.5):0.0;",
    # 6. distances and leaf names (popular)
    "(A:0.1,B:0.2,(C:0.3,D:0.4):0.5);",
    # 7. distances and all names
    "(A:0.1,B:0.2,(C:0.3,D:0.4)E:0.5)F;",
)


@pytest.mark.parametrize("nwk_str", _STR_EQUALITY)
def test_newick__wikipedia_examples__str_equality(nwk_str):
    nodes = Newick.from_string(nwk_str)
    result = str(nodes)
    assert result == nwk_str


###############################################################################
###############################################################################
# Immutability

_PROPERTIES_ARE_IMMUTABLE = (
    ("name", "foo"),
    ("length", "13"),
    ("children", []),
    ("foobar", True),
)


@pytest.mark.parametrize("name, value", _PROPERTIES_ARE_IMMUTABLE)
def test_newick__properties_are_immutable(name, value):
    node = Newick(name="A", length=3, children=[Newick(name="B")])
    with pytest.raises(NotImplementedError):
        setattr(node, name, value)


@pytest.mark.parametrize("name", ("name", "length", "children", "foobar"))
def test_newick__properties_cannot_be_deleted(name):
    node = Newick(name="A", length=3, children=[Newick(name="B")])
    with pytest.raises(NotImplementedError):
        delattr(node, name)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.common.utilities import safe_coerce_to_tuple
from paleomix.nodes.picard import MergeSamFilesNode
from paleomix.nodes.validation import DetectInputDuplicationNode
from paleomix.pipelines.bam.nodes import index_and_validate_bam


class Prefix:
    def __init__(self, config, prefix, samples, features, target):
        self.name = prefix["Name"]
        self.roi = prefix.get("RegionsOfInterest", {})

        self.samples = safe_coerce_to_tuple(samples)
        self.folder = config.destination
        self.target = target

        files_and_nodes = {}
        for sample in self.samples:
            files_and_nodes.update(sample.bams.items())

        self.datadup_check = self._build_dataduplication_node(prefix, files_and_nodes)
        self.bams = self._build_bam(config, prefix, files_and_nodes)

        nodes = [self.datadup_check]
        for sample in self.samples:
            nodes.extend(sample.nodes)
        self.nodes = tuple(nodes)

    def _build_bam(self, config, prefix, files_and_bams):
        output_filename = os.path.join(
            self.folder, "%s.%s.bam" % (self.target, prefix["Name"])
        )
        validated_filename = os.path.join(
            self.folder, self.target, prefix["Name"] + ".validated"
        )

        node = MergeSamFilesNode(
            config=config,
            input_bams=list(files_and_bams),
            output_bam=output_filename,
            dependencies=self.datadup_check,
        )
        validated_node = index_and_validate_bam(
            config=config,
            prefix=prefix,
            node=node,
            log_file=validated_filename,
            validation_levels=("full", "partial"),
        )

        return {output_filename: validated_node}

    def _build_dataduplication_node(self, prefix, files_and_nodes):
        filename = prefix["Name"] + ".duplications_checked"
        destination = os.path.join(self.folder, self.target, filename)
        dependencies = list(files_and_nodes.values())

        return DetectInputDuplicationNode(
            input_files=list(files_and_nodes),
            output_file=destination,
            dependencies=dependencies,
        )



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
import copy

import paleomix.pipelines.bam.paths as paths

from paleomix.common.makefile import MakefileError

from paleomix.nodes.bwa import BWAAlgorithmNode, BWABacktrack, BWASampe, BWASamse
from paleomix.nodes.bowtie2 import Bowtie2Node

from paleomix.pipelines.bam.parts import Reads
from paleomix.pipelines.bam.nodes import index_and_validate_bam

from paleomix.common.fileutils import swap_ext


#
_TRIMMED_READS_CACHE = {}


class Lane:
    def __init__(self, config, prefix, record, name):
        self.name = name
        self.bams = {}
        self.reads = None
        self.options = copy.deepcopy(record["Options"])
        self.tags = tags = copy.deepcopy(record["Tags"])
        self.tags["PG"] = self.tags["PG"].lower()
        self.folder = os.path.join(
            config.destination,
            tags["Target"],
            prefix["Name"],
            tags["SM"],
            tags["LB"],
            tags["PU"],
            tags["Folder"],
        )

        self._init_reads(config, record)
        self._init_unaligned_lane(config, prefix, record)

    def _init_reads(self, config, record):
        key = tuple(self.tags[key] for key in ("Target", "SM", "LB", "PU", "DS"))
        if key not in _TRIMMED_READS_CACHE:
            _TRIMMED_READS_CACHE[key] = Reads(
                config, record, record["Options"]["QualityOffset"]
            )
        self.reads = _TRIMMED_READS_CACHE[key]

    def _init_unaligned_lane(self, config, prefix, record):
        prefix_key = "Nodes:%s" % (self.options["Aligners"]["Program"],)

        for (key, input_filename) in self.reads.files.items():
            # Common parameters between BWA / Bowtie2
            output_filename = os.path.join(self.folder, "%s.bam" % (key.lower(),))

            alignment_node = self._build_alignment_node(
                config=config,
                record=record,
                prefix=prefix,
                parameters={
                    "input_file": input_filename,
                    "output_file": output_filename,
                    "prefix": prefix["Path"],
                    "reference": prefix["Reference"],
                    "dependencies": self.reads.nodes + prefix[prefix_key],
                },
            )

            self.bams[key] = {
                output_filename: self._finalize_nodes(
                    config=config, prefix=prefix, node=alignment_node,
                )
            }

    def _build_alignment_node(self, config, record, prefix, parameters):
        if self.options["Aligners"]["Program"] == "BWA":
            algorithm = self.options["Aligners"]["BWA"]["Algorithm"].lower()
            parameters["threads"] = config.bwa_max_threads
            parameters["algorithm"] = algorithm

            if algorithm == "backtrack":
                return self._build_bwa_backtrack(
                    config=config, record=record, prefix=prefix, parameters=parameters
                )
            elif algorithm in ("mem", "bwasw"):
                return self._build_bwa_algorithm(
                    config=config, record=record, prefix=prefix, parameters=parameters
                )
            else:
                raise NotImplementedError("BWA %r not implemented!" % (algorithm,))

        elif self.options["Aligners"]["Program"] == "Bowtie2":
            return self._build_bowtie2(
                config=config, record=record, prefix=prefix, parameters=parameters
            )
        else:
            raise NotImplementedError(
                "Aligner %r not implemented!" % (self.options["Aligners"]["Program"],)
            )

    ###########################################################################
    ###########################################################################
    # Construction of mapping nodes

    def _build_bwa_backtrack(self, config, prefix, record, parameters):
        if paths.is_paired_end(parameters["input_file"]):
            return self._build_bwa_backtrack_pe(
                config=config, prefix=prefix, record=record, parameters=parameters
            )
        else:
            return self._build_bwa_backtrack_se(
                config=config, prefix=prefix, record=record, parameters=parameters
            )

    def _build_bwa_backtrack_aln(self, parameters, input_file, output_file):
        options = dict(self.options["Aligners"]["BWA"])
        if not self.options["Aligners"]["BWA"]["UseSeed"]:
            options["-l"] = 2 ** 16 - 1

        if self.options["QualityOffset"] in (64, "Solexa"):
            options["-I"] = None

        return BWABacktrack(
            input_file=input_file,
            output_file=output_file,
            threads=parameters["threads"],
            prefix=parameters["prefix"],
            reference=parameters["reference"],
            mapping_options=options,
            dependencies=parameters["dependencies"],
        )

    def _build_bwa_backtrack_se(self, config, prefix, record, parameters):
        input_file_fq = parameters.pop("input_file")
        output_file_bam = parameters.pop("output_file")
        output_file_sai = swap_ext(output_file_bam, ".sai")

        aln_node = self._build_bwa_backtrack_aln(
            parameters=parameters, input_file=input_file_fq, output_file=output_file_sai
        )

        return BWASamse(
            input_file_fq=input_file_fq,
            input_file_sai=output_file_sai,
            output_file=output_file_bam,
            prefix=parameters["prefix"],
            reference=parameters["reference"],
            mapping_options=self.options["Aligners"]["BWA"],
            cleanup_options=self._cleanup_options("BWA"),
            dependencies=aln_node,
        )

    def _build_bwa_backtrack_pe(self, config, prefix, record, parameters):
        template = parameters.pop("input_file")
        output_bam = parameters.pop("output_file")

        output_sai_1 = swap_ext(output_bam, "%i.sai" % (1,))
        aln_node_1 = self._build_bwa_backtrack_aln(
            parameters=parameters,
            input_file=template.format(Pair=1),
            output_file=output_sai_1,
        )

        output_sai_2 = swap_ext(output_bam, "%i.sai" % (2,))
        aln_node_2 = self._build_bwa_backtrack_aln(
            parameters=parameters,
            input_file=template.format(Pair=2),
            output_file=output_sai_2,
        )

        return BWASampe(
            input_file_sai_1=output_sai_1,
            input_file_sai_2=output_sai_2,
            input_file_fq_1=template.format(Pair=1),
            input_file_fq_2=template.format(Pair=2),
            output_file=output_bam,
            prefix=parameters["prefix"],
            reference=parameters["reference"],
            mapping_options=self.options["Aligners"]["BWA"],
            cleanup_options=self._cleanup_options("BWA"),
            dependencies=(aln_node_1, aln_node_2),
        )

    def _build_bwa_algorithm(self, config, prefix, record, parameters):
        if self.options["QualityOffset"] != 33:
            raise MakefileError(
                "Mapping with BWA using the %r algorithm currently does not support "
                "QualityOffsets other than 33; please convert your FASTQ if you wish "
                "to proceed." % (parameters["algorithm"],)
            )

        parameters = self._set_pe_input_files(parameters)
        parameters["mapping_options"] = self.options["Aligners"]["BWA"]
        parameters["cleanup_options"] = self._cleanup_options("BWA")

        return BWAAlgorithmNode(**parameters)

    def _build_bowtie2(self, config, prefix, record, parameters):
        parameters = self._set_pe_input_files(parameters)
        parameters["threads"] = config.bowtie2_max_threads

        parameters["mapping_options"] = dict(self.options["Aligners"]["Bowtie2"])
        parameters["cleanup_options"] = self._cleanup_options("Bowtie2")

        if self.options["QualityOffset"] != 33:
            parameters["mapping_options"]["--phred64"] = None

        return Bowtie2Node(**parameters)

    def _cleanup_options(self, aligner):
        options = {
            "--rg-id": self.tags["ID"],
            "--rg": [
                "%s:%s" % (tag_name, self.tags[tag_name])
                for tag_name in ("SM", "LB", "PU", "PL", "PG", "DS")
            ],
            "-q": self.options["Aligners"][aligner]["MinQuality"],
        }

        if self.options["Aligners"][aligner]["FilterUnmappedReads"]:
            options["-F"] = "0x4"

        return options

    def _finalize_nodes(self, config, prefix, node):
        # Indexes are required for all files when calculating region statistics
        return index_and_validate_bam(
            config=config,
            prefix=prefix,
            node=node,
            create_index=bool(prefix.get("RegionsOfInterest")),
        )

    @staticmethod
    def _set_pe_input_files(parameters):
        parameters = dict(parameters)
        template = parameters.pop("input_file")
        if paths.is_paired_end(template):
            parameters["input_file_1"] = template.format(Pair=1)
            parameters["input_file_2"] = template.format(Pair=2)
        else:
            parameters["input_file_1"] = template
            parameters["input_file_2"] = None

        return parameters



#!/usr/bin/python3
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from paleomix.common.utilities import safe_coerce_to_frozenset, get_in, set_in

from paleomix.common.formats import FormatError


class GraphError(FormatError):
    pass


class _Graph:
    """Internal representation of an unrooted graph, allowing various forms of
    manipulation directly on the graph. To ensure that all manipulations can be
    carried out, it is required that branch-lengths are present for ALL branches,
    or for NO branches.

    Note that neither the root-length, nor node-ordering is preserved."""

    def __init__(self):
        self.names = {}
        self.connections = {}
        self.has_branch_lengths = None

    def is_leaf(self, node):
        """Returns true if the node is a leaf, defined as having a single connection."""
        return len(self.connections[node]) == 1

    def get_path_length(self, *nodes):
        """Returns the length of a path through the graph. Calling the function
        with two nodes is the equivalent of getting the branch-length between
        those two nodes."""
        if not self.has_branch_lengths:
            return None

        path_length = 0.0
        for (node_a, node_b) in zip(nodes, nodes[1:]):
            segment_length = float(self.connections[node_a][node_b])
            path_length += segment_length

        return path_length

    def set_name(self, node_id, name):
        self.names[node_id] = name

    def add_connection(self, node_id_a, node_id_b, blength=None):
        if (blength is not None) and float(blength) < 0:
            raise GraphError("Branch-lengths must be non-negative")
        elif (blength is not None) != self.has_branch_lengths:
            if self.has_branch_lengths is not None:
                raise GraphError("Tree contains branches with and without lengths")
            self.has_branch_lengths = blength is not None

        set_in(self.connections, (node_id_a, node_id_b), blength)
        set_in(self.connections, (node_id_b, node_id_a), blength)

    def remove_connection(self, node_a, node_b):
        length_a = self.connections[node_a].pop(node_b)
        length_b = self.connections[node_b].pop(node_a)
        assert length_a == length_b, (length_a, length_b)
        return length_a

    def remove_node(self, node):
        connections = self.connections.pop(node)
        for node_b in connections:
            self.connections[node_b].pop(node)
        self.names.pop(node)

    def rebuild_tree(self, parent_id, node_id):
        """Rebuilds a tree starting at a node with id
        'node_id' and a parent with id 'parent_id' (or the
        same value as 'node_id' if a root node)."""
        raise NotImplementedError(
            "Subclasses must implement 'rebuild_nodes'."
        )  # pragma: no coverage

    def prune_uninformative_nodes(self):
        """Removes nodes without names, and which are connected
        to two other nodes, extending the branch lengths of the
        two connected nodes. This process is repreated, until no
        further nodes are pruned. A rooted tree will typically
        contain just 1 such node, namely the old root node.

        For example, the tree "(A:5,(B:6):3);" would be reduced to
        the tree "(A:5,B:9);", whereas the trees "(A:5,(B:6)C:3);"
        and "(A:5,(B:6,C:2):3);" would not be pruned.

        For a node to be pruned, both adjacent nodes must have a
        length specified, or both must not have a length specified."""
        while True:
            for (cur_node, connections) in self.connections.items():
                if not self.names[cur_node] and (len(connections) == 2):
                    conn_a, conn_b = connections

                    blength = self.get_path_length(conn_a, cur_node, conn_b)

                    # Splice out the current node
                    self.remove_node(cur_node)
                    self.add_connection(conn_a, conn_b, blength)
                    break
            else:
                # Nothing was pruned this round, terminate
                break

    ################################################################################
    ################################################################################
    # Functions relating to NEWICK rooting on midpoint

    def reroot_on_midpoint(self):
        if not self.has_branch_lengths:
            raise GraphError(
                "Cannot reroot on midpoint for tree without branch-lengths"
            )

        longest_path, length = self._find_longest_path()
        root = self._create_root_at(longest_path, length / 2.0)

        return self.rebuild_tree(root, root)

    def _find_longest_path(self):
        """This function determines the longest non-overlapping path possible,
        and returns a list of the sequence of nodes in this path, as well as
        the total length of this path."""
        path_blengths = {}
        path_guides = {}

        def _collect_paths(guide, length, p_node, c_node):
            length += self.get_path_length(p_node, c_node)

            guide.append(c_node)
            key = frozenset(guide)
            path_blengths[key] = length
            path_guides[key] = guide

            for other in self.connections[c_node]:
                if other not in key:
                    _collect_paths(list(guide), length, c_node, other)

        for (p_node, connections) in self.connections.items():
            for c_node in connections:
                _collect_paths([p_node], 0, p_node, c_node)

        key, length = max(path_blengths.items(), key=lambda item: item[1])
        return path_guides[key], length

    def _create_root_at(self, path, root_at):
        """Finds the midpoint of a path through a tree, and
        either creates a new node at that point, or selects
        the node already present at that point (if any). The
        mid-point is assumed to be at distance of 'root_at'
        from the starting node.

        E.g. if the path is the longest path, and 'root_at' is
        half the length of this path, then this corresponds to
        rooting at the midpoint.

        The id of the new / selected node is returned. New
        nodes (if created) are always given the id None."""
        for (c_node, n_node) in zip(path, path[1:]):
            branch_length = self.get_path_length(c_node, n_node)

            if branch_length > root_at:
                left_len = root_at
                right_len = branch_length - root_at

                self.remove_connection(c_node, n_node)
                self.add_connection(None, c_node, left_len)
                self.add_connection(None, n_node, right_len)

                return None
            elif branch_length == root_at:
                return n_node
            root_at -= branch_length
        assert False  # pragma: no coverage

    ################################################################################
    ################################################################################
    # Functions relating to NEWICK rooting on taxa

    def reroot_on_taxa(self, taxa):
        taxa = safe_coerce_to_frozenset(taxa)
        if not taxa:
            raise ValueError("No taxa in outgroup")

        clades = self._collect_clades()
        root_on = self._collect_nodes_from_names(taxa)
        # Because None is the id of the root atm:
        root = self._create_root_with_clade(clades, root_on)

        return self.rebuild_tree(root, root)

    def _collect_nodes_from_names(self, taxa):
        known_taxa = set()
        for (node_id, name) in self.names.items():
            if self.is_leaf(node_id):
                known_taxa.add(name)

        unknown_taxa = taxa - known_taxa
        if unknown_taxa:
            raise ValueError(
                "Cannot root on unknown taxa: %s" % (", ".join(unknown_taxa),)
            )
        elif not (known_taxa - taxa):
            raise ValueError("Cannot root on every taxa in tree")

        return frozenset(key for (key, name) in self.names.items() if name in taxa)

    def _collect_clades(self):
        clades = {}
        for (node_a, connections) in self.connections.items():
            for node_b in connections:
                self._collect_clade_from(clades, node_a, node_b)
        return clades

    def _collect_clade_from(self, cache, p_node, c_node):
        c_clade = get_in(cache, (p_node, c_node), set())
        if not c_clade:
            if self.is_leaf(c_node):
                c_clade.add(c_node)

            for n_node in self.connections[c_node]:
                if n_node != p_node:
                    c_clade.update(self._collect_clade_from(cache, c_node, n_node))
            set_in(cache, (p_node, c_node), frozenset(c_clade))
        return c_clade

    def _create_root_with_clade(self, clades, taxa):
        root_key, root_clade, root_length = None, None, None
        for (p_node, connections) in clades.items():
            for (n_node, clade) in connections.items():
                if (root_clade is None) or (len(clade) < len(root_clade)):
                    if taxa.issubset(clade):
                        root_key = (p_node, n_node)
                        root_clade = clade
                        root_length = self.get_path_length(p_node, n_node)

        p_node, n_node = root_key
        if root_length is not None:
            root_length = float(root_length) / 2.0

        self.remove_connection(p_node, n_node)
        self.add_connection(None, p_node, root_length)
        self.add_connection(None, n_node, root_length)

        return None

    ################################################################################
    ################################################################################
    # Functions relating to calculating bootstrap support
    def get_clade_names(self):
        result = set()
        for (_, connections) in self._collect_clades().items():
            for (_, clade) in connections.items():
                result.add(frozenset(self.names[node_id] for node_id in clade))
        return result



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import bz2
import gzip
import io

import pytest

from paleomix.common.fileutils import fspath
from paleomix.common.formats.fasta import FASTA, FASTAError


_SEQ_FRAG = "AAGTCC"  # len() = 6


###############################################################################
###############################################################################
# Tests for FASTA constructor


def _simple_fasta_record():
    return FASTA("Dummy", "Meta-inf", "ACGT")


def test_fasta__constructor__name():
    record = _simple_fasta_record()
    assert record.name == "Dummy"


def test_fasta__constructor__meta():
    record = _simple_fasta_record()
    assert record.meta == "Meta-inf"


def test_fasta__constructor__sequence():
    record = _simple_fasta_record()
    assert record.sequence == "ACGT"


def test_fasta__constructor__name_must_be_non_empty():
    with pytest.raises(FASTAError):
        FASTA("", None, "ACGT")


def test_fasta__constructor__name_must_be_string_type():
    with pytest.raises(FASTAError):
        FASTA(1, None, "ACGT")


def test_fasta__constructor__name_must_be_string_type_or_none():
    with pytest.raises(FASTAError):
        FASTA("Seq1", 1, "ACGT")


def test_fasta__constructor__sequence_must_be_string_type():
    with pytest.raises(FASTAError):
        FASTA("Seq1", None, 1)


###############################################################################
###############################################################################
# Tests for __repr__


def test_fasta__repr__partial_line_test():
    expected = "FASTA('foobar', '', %r)" % (_SEQ_FRAG,)
    result = repr(FASTA("foobar", None, _SEQ_FRAG))
    assert result == expected


def test_fasta__repr__complete_line_test():
    expected = "FASTA('barfoo', '', %r)" % (_SEQ_FRAG * 10,)
    result = repr(FASTA("barfoo", None, _SEQ_FRAG * 10))
    assert result == expected


def test_fasta__repr__multiple_lines():
    expected = "FASTA('foobar', '', %r)" % (_SEQ_FRAG * 15,)
    result = repr(FASTA("foobar", None, _SEQ_FRAG * 15))
    assert result == expected


def test_fasta__repr__partial_line_test_with_meta_information():
    expected = "FASTA('foobar', 'my Meta-Info', %r)" % (_SEQ_FRAG,)
    result = repr(FASTA("foobar", "my Meta-Info", _SEQ_FRAG))
    assert result == expected


###############################################################################
###############################################################################
# Tests for write


def test_fasta__write__partial_line():
    expected = ">foobar\n%s\n" % (_SEQ_FRAG,)
    stringf = io.StringIO()
    FASTA("foobar", None, _SEQ_FRAG).write(stringf)
    assert stringf.getvalue() == expected


def test_fasta__write__with_metadata():
    expected = ">foobar my Meta data\n%s\n" % (_SEQ_FRAG,)
    stringf = io.StringIO()
    FASTA("foobar", "my Meta data", _SEQ_FRAG).write(stringf)
    assert stringf.getvalue() == expected


def test_fasta__write__complete_line_test():
    expected = ">barfoo\n%s\n" % (_SEQ_FRAG * 10,)
    stringf = io.StringIO()
    FASTA("barfoo", None, _SEQ_FRAG * 10).write(stringf)
    assert stringf.getvalue() == expected


def test_fasta__write__multiple_lines():
    expected = ">foobar\n%s\n%s\n" % (_SEQ_FRAG * 10, _SEQ_FRAG * 5)
    stringf = io.StringIO()
    FASTA("foobar", None, _SEQ_FRAG * 15).write(stringf)
    assert stringf.getvalue() == expected


###############################################################################
###############################################################################
# Tests for FASTA.from_lines


def test_fasta__from_lines__no_records():
    assert list(FASTA.from_lines([])) == list([])


def test_fasta__from_lines_single_record():
    lines = [">single\n", "TGTTCTCCACCGTGCACAAC\n", "CCTTCATCCA\n"]
    expected = [FASTA("single", None, "TGTTCTCCACCGTGCACAACCCTTCATCCA")]
    assert list(FASTA.from_lines(lines)) == list(expected)


def test_fasta__from_lines__multiple_records():
    lines = [
        ">first\n",
        "TGTTCTCCACCGTGCACAAC\n",
        "CCTTCATCCA\n",
        ">Second XT:1:0\n",
        "GAGAGCTCAGCTAAC\n",
        ">Third\n",
        "CGCTGACCAAAAACGGACAG\n",
        "GGCATTCGGC\n",
    ]
    expected = [
        FASTA("first", None, "TGTTCTCCACCGTGCACAACCCTTCATCCA"),
        FASTA("Second", "XT:1:0", "GAGAGCTCAGCTAAC"),
        FASTA("Third", None, "CGCTGACCAAAAACGGACAGGGCATTCGGC"),
    ]
    assert list(FASTA.from_lines(lines)) == list(expected)


def test_fasta__from_lines__empty_record_name_only__nothing_else():
    with pytest.raises(FASTAError):
        list(FASTA.from_lines([">fasta1\n"]))


def test_fasta__from_lines__empty_record_name_only__first():
    with pytest.raises(FASTAError):
        list(FASTA.from_lines([">fasta1\n", ">fasta2\n", "AGTC\n"]))


def test_fasta__from_lines__empty_record__middle():
    lines = [">fasta0\n", "ACGT\n", ">fasta1\n", ">fasta2\n", "AGTC\n"]
    with pytest.raises(FASTAError):
        list(FASTA.from_lines(lines))


def test_fasta__from_lines__empty_record_last():
    lines = [">fasta1\n", "ACGT\n", ">fasta2\n"]
    with pytest.raises(FASTAError):
        list(FASTA.from_lines(lines))


def test_fasta__from_lines__missing_name__alone():
    with pytest.raises(FASTAError):
        list(FASTA.from_lines(["ACGT\n"]))


def test_fasta__from_lines__missing_name__with_others():
    with pytest.raises(FASTAError):
        list(FASTA.from_lines(["ACGT\n", ">Foo\n", "ACGGTA\n"]))


def test_fasta__from_lines__empty_name__alone():
    with pytest.raises(FASTAError):
        list(FASTA.from_lines([">\n", "ACGT\n"]))


def test_fasta__from_lines__empty_name__with_others():
    with pytest.raises(FASTAError):
        list(FASTA.from_lines([">\n", "ACGT\n", ">Foo\n", "ACGGTA\n"]))


###############################################################################
###############################################################################
# Tests for 'FASTA.from_file'


@pytest.mark.parametrize("func", (open, gzip.open, bz2.open))
def test_fasta__from_file(func, tmp_path):
    expected = [
        FASTA("This_is_FASTA!", None, "ACGTN"),
        FASTA("This_is_ALSO_FASTA!", None, "CGTNA"),
    ]

    with func(fspath(tmp_path / "file"), "wt") as handle:
        for item in expected:
            item.write(handle)

    assert list(FASTA.from_file(tmp_path / "file")) == expected


###############################################################################
###############################################################################


def test_fasta__equality():
    assert FASTA("A", "B", "C") == FASTA("A", "B", "C")


def test_fasta__inequality():
    assert FASTA("A", "B", "C") != FASTA("A", "B", "D")
    assert FASTA("A", "B", "C") != FASTA("A", None, "C")
    assert FASTA("A", "B", "C") != FASTA("D", "B", "C")


def test_fasta__sorting_less_equal():
    assert not FASTA("A", "B", "C") < FASTA("A", "B", "C")
    assert FASTA("A", "B", "C") < FASTA("B", "B", "C")
    assert FASTA("A", "B", "C") < FASTA("A", "C", "C")
    assert FASTA("A", "B", "C") < FASTA("A", "B", "D")
    assert FASTA("A", "B", "C") <= FASTA("A", "B", "C")
    assert FASTA("A", "B", "C") <= FASTA("B", "B", "C")
    assert FASTA("A", "B", "C") <= FASTA("A", "C", "C")
    assert FASTA("A", "B", "C") <= FASTA("A", "B", "D")


def test_fasta__sorting_greater_equal():
    assert not FASTA("A", "B", "C") > FASTA("A", "B", "C")
    assert FASTA("B", "B", "C") > FASTA("A", "B", "C")
    assert FASTA("A", "C", "C") > FASTA("A", "B", "C")
    assert FASTA("A", "B", "D") > FASTA("A", "B", "C")
    assert FASTA("A", "B", "C") >= FASTA("A", "B", "C")
    assert FASTA("B", "B", "C") >= FASTA("A", "B", "C")
    assert FASTA("A", "C", "C") >= FASTA("A", "B", "C")
    assert FASTA("A", "B", "D") >= FASTA("A", "B", "C")


def test_fasta__hash():
    assert hash(FASTA("A", "B", "C")) == hash(FASTA("A", "B", "C"))
    assert hash(FASTA("A", "B", "C")) != hash(FASTA("B", "B", "C"))
    assert hash(FASTA("A", "B", "C")) != hash(FASTA("A", "C", "C"))
    assert hash(FASTA("A", "B", "C")) != hash(FASTA("A", "B", "D"))


def test_fasta__unimplemented_comparison():
    assert NotImplemented is FASTA("A", None, "C").__eq__(10)
    assert NotImplemented is FASTA("A", None, "C").__lt__(10)
    assert NotImplemented is FASTA("A", None, "C").__le__(10)
    assert NotImplemented is FASTA("A", None, "C").__ge__(10)
    assert NotImplemented is FASTA("A", None, "C").__gt__(10)


###############################################################################
###############################################################################
# Tests for index_and_collect_contigs

_TEST_FASTA_1_A = FASTA("seq1", "meta1", "TCTTTCAGTCTGGAGACTAGCCTCC")
_TEST_FASTA_1_B = FASTA("seq1", None, "ATTGAGGCGTATTGTGTCG")
_TEST_FASTA_2 = FASTA("seq2", None, "CAAAGCA")
_TEST_FASTA_3 = FASTA("seq3", "more meta data", "AGCTCTCCTCCCC")


def test_index_and_collect_contigs(tmp_path):
    fasta_file = tmp_path / "test.fasta"
    with fasta_file.open("wt") as handle:
        _TEST_FASTA_1_A.write(handle)
        _TEST_FASTA_2.write(handle)
        _TEST_FASTA_3.write(handle)

    assert FASTA.index_and_collect_contigs(fasta_file) == {
        "seq1": 25,
        "seq2": 7,
        "seq3": 13,
    }


def test_index_and_collect_contigs__duplicate_names(tmp_path):
    fasta_file = tmp_path / "test.fasta"
    with fasta_file.open("wt") as handle:
        _TEST_FASTA_1_A.write(handle)
        _TEST_FASTA_2.write(handle)
        _TEST_FASTA_1_B.write(handle)

    assert FASTA.index_and_collect_contigs(fasta_file) == {
        "seq1": 25,
        "seq2": 7,
    }


def test_index_and_collect_contigs__fai_files(tmp_path):
    fasta_file = tmp_path / "test.fasta"
    with fasta_file.open("wt") as handle:
        _TEST_FASTA_1_A.write(handle)

    fai_file = tmp_path / "test.fasta.fai"

    # Fai file should be created once, and then not modified
    FASTA.index_and_collect_contigs(fasta_file)
    stats_1 = fai_file.stat()
    FASTA.index_and_collect_contigs(fasta_file)
    stats_2 = fai_file.stat()

    assert stats_1 == stats_2



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from paleomix.node import CommandNode, NodeError
from paleomix.atomiccmd.builder import (
    AtomicCmdBuilder,
    apply_options,
)
from paleomix.common.fileutils import reroot_path
from paleomix.common.formats.msa import MSA, MSAError
import paleomix.common.versions as versions


MAFFT_VERSION = versions.Requirement(
    call=("mafft", "--version"), search=r"v(\d+)\.(\d+)", checks=versions.GE(7, 307)
)


# Presets mainly taken from
# http://mafft.cbrc.jp/alignment/software/algorithms/algorithms.html
_PRESETS = {
    "mafft": ["mafft"],
    "auto": ["mafft", "--auto"],
    "fft-ns-1": ["mafft-fftns", "--retree", 1],
    "fft-ns-2": ["mafft-fftns"],
    "fft-ns-i": ["mafft-fftnsi"],
    "nw-ns-i": ["mafft-nwnsi"],
    "l-ins-i": ["mafft-linsi"],
    "e-ins-i": ["mafft-einsi"],
    "g-ins-i": ["mafft-ginsi"],
}


class MAFFTNode(CommandNode):
    def __init__(
        self, input_file, output_file, algorithm="auto", options={}, dependencies=()
    ):
        command = AtomicCmdBuilder(
            _PRESETS[algorithm.lower()],
            IN_FASTA=input_file,
            OUT_STDOUT=output_file,
            CHECK_VERSION=MAFFT_VERSION,
        )

        apply_options(command, options)

        # Needs to be specified after options and flags
        command.add_value("%(IN_FASTA)s")

        self._output_file = output_file

        CommandNode.__init__(
            self,
            command=command.finalize(),
            description="aligning sequences in %s using MAFFT %s"
            % (input_file, algorithm),
            dependencies=dependencies,
        )

    def _teardown(self, config, temp):
        # Validate output from MAFFT
        output_file = reroot_path(temp, self._output_file)
        try:
            MSA.from_file(output_file)
        except MSAError as error:
            raise NodeError("Invalid MSA produced by MAFFT:\n%s" % (error,))
        CommandNode._teardown(self, config, temp)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import itertools
import re


class TableError(RuntimeError):
    pass


_MIN_PADDING = 4
_WHITESPACE_OR_EMPTY = re.compile(r"\s|^$")


def padded_table(table):
    """Takes a sequence of iterables, each of which represents a row in a
    table. Values are converted to string, and padded with whitespace such that
    each column is separated from its adjacent columns by at least 4 spaces.
    Empty cells or whitespace in values are not allowed.

    If a string is included instead of a row, this value is added as is. Note
    that these lines should be whitespace only, or start with a '#' if the
    resulting table is to be readable with 'parse_padded_table'.
    """
    str_rows = []
    max_sizes = []
    for row in table:
        if not isinstance(row, str):
            row = list(map(str, row))
            row_sizes = list(map(len, row))
            max_sizes = list(
                map(max, itertools.zip_longest(max_sizes, row_sizes, fillvalue=0))
            )

        str_rows.append(row)

    sizes = [(size + _MIN_PADDING) for size in max_sizes]
    for row in str_rows:
        if not isinstance(row, str):
            row = "".join(
                field.ljust(padding) for (field, padding) in zip(row, sizes)
            ).rstrip()
        yield row


def parse_padded_table(lines, header=None):
    """Parses a padded table generated using 'padded_table', or any table which
    consists of a fixed number of columns seperated by whitespace, with no
    whitespace in the cells. Empty lines and lines starting with '#' (comments)
    are ignored. Each row is returned as a dictionary, using the values found
    in the first row as keys.
    """
    for line in lines:
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        elif header is None:
            header = stripped.split()
            nheader = len(header)
            continue

        fields = stripped.split()
        if len(fields) != nheader:
            raise TableError(
                "Malformed table; #columns does not match header:"
                " %r vs %r" % (header, fields)
            )

        yield dict(zip(header, fields))


def parse_lines(lines, parser):
    """Parses a set of lines using the supplied callable:
        lambda (line, length): ...

    Supports the parser functions available in 'pysam': asGTF, asBED, etc.
    """
    if not callable(parser):
        raise TypeError(
            "'parser' must be a callable, not %r" % parser.__class__.__name__
        )

    for line in lines:
        stripped = line.lstrip()
        if stripped and stripped[0] not in ("#", 35):
            stripped = line.rstrip()
            yield parser(stripped, len(stripped))


def parse_lines_by_contig(lines, parser):
    """Reads the lines of a text file, parsing each line with the specified
    parser, and aggregating results by the 'contig' property of reach record.
    """
    table = {}
    for record in parse_lines(lines, parser):
        try:
            table[record.contig].append(record)
        except KeyError:
            table[record.contig] = [record]

    return table



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import functools
import os

import paleomix.common.versions as versions
import paleomix.tools.factory as factory

from paleomix.atomiccmd.builder import (
    AtomicCmdBuilder,
    apply_options,
)
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.atomiccmd.sets import ParallelCmds
from paleomix.common.fileutils import describe_paired_files
from paleomix.node import CommandNode, NodeError
from paleomix.nodes.samtools import SAMTOOLS_VERSION


BWA_VERSION = versions.Requirement(
    name="BWA",
    call=("bwa",),
    search=r"Version: (\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(0, 7, 15),
)


class BWAIndexNode(CommandNode):
    def __init__(self, input_file, prefix=None, dependencies=()):
        prefix = prefix if prefix else input_file
        builder = _new_bwa_command(
            ("bwa", "index", "%(IN_FILE)s", "-p", "%(TEMP_OUT_PREFIX)s"),
            prefix,
            iotype="OUT",
            IN_FILE=input_file,
            TEMP_OUT_PREFIX=os.path.basename(prefix),
        )

        CommandNode.__init__(
            self,
            command=builder.finalize(),
            description="creating BWA index for %s" % (input_file,),
            dependencies=dependencies,
        )


class BWABacktrack(CommandNode):
    def __init__(
        self,
        input_file,
        output_file,
        reference,
        prefix,
        threads=1,
        mapping_options={},
        dependencies=(),
    ):
        threads = _get_max_threads(reference, threads)

        aln = _new_bwa_command(
            ("bwa", "aln"), prefix, IN_FILE=input_file, OUT_STDOUT=output_file,
        )
        aln.add_value(prefix)
        aln.add_value("%(IN_FILE)s")
        aln.set_option("-t", threads)

        apply_options(aln, mapping_options)

        description = _get_node_description(
            name="BWA",
            algorithm="Backtrack",
            input_files_1=input_file,
            prefix=prefix,
            threads=threads,
        )

        CommandNode.__init__(
            self,
            command=aln.finalize(),
            description=description,
            threads=threads,
            dependencies=dependencies,
        )


class BWASamse(CommandNode):
    def __init__(
        self,
        input_file_fq,
        input_file_sai,
        output_file,
        reference,
        prefix,
        mapping_options={},
        cleanup_options={},
        dependencies=(),
    ):
        samse = _new_bwa_command(
            ("bwa", "samse"),
            prefix,
            IN_FILE_SAI=input_file_sai,
            IN_FILE_FQ=input_file_fq,
            OUT_STDOUT=AtomicCmd.PIPE,
        )
        samse.add_value(prefix)
        samse.add_value("%(IN_FILE_SAI)s")
        samse.add_value("%(IN_FILE_FQ)s")

        cleanup = _new_cleanup_command(samse, output_file, reference)

        apply_options(samse, mapping_options)
        apply_options(cleanup, cleanup_options)

        CommandNode.__init__(
            self,
            command=ParallelCmds([samse.finalize(), cleanup.finalize()]),
            description=_get_node_description(
                name="BWA",
                algorithm="samse",
                input_files_1=input_file_fq,
                prefix=prefix,
            ),
            dependencies=dependencies,
        )


class BWASampe(CommandNode):
    def __init__(
        self,
        input_file_fq_1,
        input_file_fq_2,
        input_file_sai_1,
        input_file_sai_2,
        output_file,
        reference,
        prefix,
        mapping_options={},
        cleanup_options={},
        dependencies=(),
    ):
        sampe = _new_bwa_command(
            (
                "bwa",
                "sampe",
                prefix,
                "%(IN_SAI_1)s",
                "%(IN_SAI_2)s",
                "%(IN_FQ_1)s",
                "%(IN_FQ_2)s",
            ),
            prefix,
            IN_SAI_1=input_file_sai_1,
            IN_SAI_2=input_file_sai_2,
            IN_FQ_1=input_file_fq_1,
            IN_FQ_2=input_file_fq_2,
            OUT_STDOUT=AtomicCmd.PIPE,
        )

        cleanup = _new_cleanup_command(sampe, output_file, reference, paired_end=True)

        apply_options(sampe, mapping_options)
        apply_options(cleanup, cleanup_options)

        CommandNode.__init__(
            self,
            command=ParallelCmds([sampe.finalize(), cleanup.finalize()]),
            description=_get_node_description(
                name="BWA",
                algorithm="sampe",
                input_files_1=input_file_fq_1,
                input_files_2=input_file_fq_2,
                prefix=prefix,
            ),
            dependencies=dependencies,
        )


class BWAAlgorithmNode(CommandNode):
    def __init__(
        self,
        input_file_1,
        output_file,
        reference,
        prefix,
        input_file_2=None,
        threads=1,
        algorithm="mem",
        mapping_options={},
        cleanup_options={},
        dependencies=(),
    ):
        if algorithm not in ("mem", "bwasw"):
            raise NotImplementedError("BWA algorithm %r not implemented" % (algorithm,))

        threads = _get_max_threads(reference, threads)

        aln = _new_bwa_command(
            ("bwa", algorithm, prefix, "%(IN_FILE_1)s"),
            prefix,
            IN_FILE_1=input_file_1,
            OUT_STDOUT=AtomicCmd.PIPE,
        )

        if input_file_2:
            aln.add_value("%(IN_FILE_2)s")
            aln.set_kwargs(IN_FILE_2=input_file_2)

        aln.set_option("-t", threads)
        # Mark alternative hits as secondary; required by e.g. Picard
        aln.set_option("-M")

        cleanup = _new_cleanup_command(
            aln, output_file, reference, paired_end=input_file_1 and input_file_2
        )

        apply_options(aln, mapping_options)
        apply_options(cleanup, cleanup_options)

        description = _get_node_description(
            name="BWA",
            algorithm="%s (%s)" % (algorithm, "PE" if input_file_2 else "SE"),
            input_files_1=input_file_1,
            input_files_2=input_file_2,
            prefix=prefix,
            threads=threads,
        )

        CommandNode.__init__(
            self,
            command=ParallelCmds([aln.finalize(), cleanup.finalize()]),
            description=description,
            threads=threads,
            dependencies=dependencies,
        )


def _new_cleanup_command(stdin, output_file, reference, paired_end=False):
    convert = factory.new("cleanup")
    convert.set_option("--fasta", "%(IN_FASTA_REF)s")
    convert.set_option("--temp-prefix", "%(TEMP_OUT_PREFIX)s")
    convert.set_kwargs(
        IN_STDIN=stdin,
        IN_FASTA_REF=reference,
        OUT_STDOUT=output_file,
        TEMP_OUT_PREFIX="bam_cleanup",
        CHECK_SAMTOOLS=SAMTOOLS_VERSION,
    )

    if paired_end:
        convert.set_option("--paired-end")

    return convert


def _new_bwa_command(call, prefix, iotype="IN", **kwargs):
    _check_bwa_prefix(prefix)

    kwargs["CHECK_BWA"] = BWA_VERSION
    for postfix in ("amb", "ann", "bwt", "pac", "sa"):
        kwargs["%s_PREFIX_%s" % (iotype, postfix.upper())] = prefix + "." + postfix

    return AtomicCmdBuilder(call, **kwargs)


@functools.lru_cache()
def _get_max_threads(reference, threads):
    """Returns the maximum number of threads to use when mapping against a
    given reference sequence. This is done since very little gain is obtained
    when using multiple threads for a small genome (e.g. < 1MB). If the
    reference falls below this size, only 1 thread is used (returned),
    otherwise the requested number of threads is returned.
    """
    if os.path.exists(reference) and os.path.getsize(reference) < 2 ** 20:
        return 1

    return threads


@functools.lru_cache()
def _check_bwa_prefix(prefix):
    """Checks that a given prefix is compatible with the currently required version of
    BWA. Older index files are incompatible with BWA v0.7.x, but may be identified by
    the presense of a small number of additional files not present when an index is
    produced using BWA v0.7.x.
    """
    if any(os.path.exists(prefix + ext) for ext in (".rbwt", ".rpac", ".rsa.")):
        filenames = "\n".join(
            "    %s.%s" % (prefix, ext)
            for ext in ("amb", "ann", "bwt", "pac", "sa", "rbwt", "rpac", "rsa")
        )

        raise NodeError(
            "Prefix appears to be created using BWA v0.5.x or older, but PALEOMIX only "
            "supports BWA v0.7.x or later.\nPlease remove the following files to allow "
            "PALEOMIX to re-index the FASTA file:\n%s" % (filenames,)
        )


def _get_node_description(
    name, input_files_1, input_files_2=None, algorithm=None, prefix=None, threads=1
):
    prefix = os.path.basename(prefix)
    if prefix.endswith(".fasta") or prefix.endswith(".fa"):
        prefix = prefix.rsplit(".", 1)[0]

    input_files = describe_paired_files(input_files_1, input_files_2 or ())
    info = ["alignment of", input_files, "onto", prefix, "using", name]
    if algorithm is not None:
        info.append(algorithm)

    if threads > 1:
        info.append("using %i threads" % (threads,))

    return " ".join(info)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Creats consensus sequence from a VCF and BED file:
vcf_to_fasta --intervals PATH_TO.bed --genotype PATH_TO.vcf

The VCF file is expected to have been created using SAMTools, and makes use
of SAMTools specific fields. For each region in the BED file, the script will
create a consensus sequence containing all basepairs which are filtered as
'PASS' or '.' in the VCF file; at each site, the most likely phenotype is
selected using the PL fields, and heterozygous sites encoded using UIPAC codes.

BED features with the same name are merged into single sequences. To ensure
that indels are called near sequence termini, the script expects the VCF file
to contain a certain amount of padding around the regions of interest.

"""


import copy
import itertools
import os
import sys
import re

import pysam

import paleomix.common.argparse as argparse
import paleomix.common.sequences as sequences
import paleomix.common.text as text
import paleomix.common.utilities as utilities
import paleomix.common.vcfwrap as vcfwrap

from paleomix.common.bedtools import BEDRecord


# Max number of positions to keep in memory / genotype at once
_SEQUENCE_CHUNK = 1024 * 1024  # 1kbp
# Number of columns per line in FASTA sequences
_FASTA_COLUMNS = 60

_VCF_DICT = re.compile("##(.*)=<(.*)>")


###############################################################################
###############################################################################
# Utility functions


def flush_fasta(sequence):
    """Takes a FASTA sequence as a string, fragments it into lines of exactly
    _FASTA_COLUMNS chars (e.g. 60), and prints all complete lines. The final
    incomplete line (if any) is returned.

    """
    for seq_frag in utilities.fragment(_FASTA_COLUMNS, sequence):
        if len(seq_frag) < _FASTA_COLUMNS:
            return seq_frag
        print(seq_frag)
    return ""


def split_beds(beds, size=_SEQUENCE_CHUNK):
    """Takes a list of beds, and splits each bed into chunks that are at most
    'size' bp long. The resulting (smaller) beds are returned as a new list.

    """
    results = []
    for bed in beds:
        for start in range(bed.start, bed.end, size):
            end = min(start + size, bed.end)
            split_bed = copy.copy(bed)
            split_bed.start = start
            split_bed.end = end
            results.append(split_bed)
    return results


###############################################################################
###############################################################################
# Genotyping functions


def add_snp(options, snp, position, sequence):
    if snp.alt != ".":
        genotype = "".join(vcfwrap.get_ml_genotype(snp, options.nth_sample))
        encoded = sequences.encode_genotype(genotype)
    else:
        encoded = snp.ref
    sequence[position] = encoded


def add_indel(options, bed, indel, sequence):
    if indel.alt == ".":
        return

    genotype = vcfwrap.get_ml_genotype(indel, options.nth_sample)
    if genotype[0] != genotype[1]:
        # No way to represent heterozygous indels
        return
    elif genotype[0] == "N":
        # No most likely genotype
        return

    # Note that bed.end is a past-the-end coordinate
    start = max(0, bed.start - options.padding)

    # FIXME: parse_indel only supports a single 'alt' values
    indel.alt = genotype[0]
    indel = vcfwrap.parse_indel(indel)
    if indel.in_reference:
        del_start = max(indel.pos + 1, bed.start)
        del_end = min(indel.pos + 1 + len(indel.what), bed.end)

        if del_start >= del_end:
            # Deletion does not cover any bases of interest
            return
        elif options.whole_codon_indels_only:
            if (del_end - del_start) % 3:
                # Non-codon sized overlap with area of interest
                return

        for position in range(del_start, del_end):
            sequence[position - start] = ""
    elif (len(indel.what) % 3 == 0) or not options.whole_codon_indels_only:
        # parse_indel assumes that the insertion is always the first possible
        # base when multiple positions are possible. As a consequence, the
        # position may be before start, with the rest of the bases overlapping
        # the current sequence. For example:
        #  ref = ATTT
        #  alt = ATTTT
        # It is assumed that the insertion (_) happened thus:
        #  interpretation = A_TTT
        if indel.pos >= start:
            sequence[indel.pos - start] += indel.what


def filter_vcfs(genotype, contig, start, end):
    if contig in genotype.contigs:
        parser = pysam.asVCF()
        # This raises a ValueError if the VCF does not
        # contain any entries for the specified contig.
        for vcf in genotype.fetch(contig, start, end, parser=parser):
            if vcf.filter in ("PASS", "."):
                yield vcf


def build_region(options, genotype, bed):
    # Note that bed.end is a past-the-end coordinate
    start = max(0, bed.start - options.padding)

    indels = []
    sequence = ["N"] * (bed.end - start)
    for vcf in filter_vcfs(genotype, bed.contig, start, bed.end):
        if vcfwrap.is_indel(vcf):
            indels.append(vcf)
        else:
            add_snp(options, vcf, vcf.pos - start, sequence)

    if not options.ignore_indels:
        for vcf in indels:
            add_indel(options, bed, vcf, sequence)

    offset = bed.start - start
    length = bed.end - bed.start
    truncated = sequence[offset : offset + length]

    # Discard insertions after the last position
    truncated[-1] = truncated[-1][:1]

    return "".join(truncated)


def build_regions(options, genotype, beds, reverse_compl):
    for bed in beds:
        sequence = build_region(options, genotype, bed)
        if reverse_compl:
            sequence = sequences.reverse_complement(sequence)
        yield sequence


def build_genes(options, genotype, regions):
    def keyfunc(bed):
        return (bed.contig, bed.name, bed.start)

    regions.sort(key=keyfunc)

    for (gene, beds) in itertools.groupby(regions, lambda x: x.name):
        beds = split_beds(beds)
        reverse_compl = False
        if any((bed.strand == "-") for bed in beds):
            assert all((bed.strand == "-") for bed in beds)

            beds.reverse()
            reverse_compl = True

        fragments = build_regions(options, genotype, beds, reverse_compl)
        yield (gene, fragments)


def genotype_genes(options, intervals, genotype):
    for (_, beds) in sorted(intervals.items()):
        for (name, fragments) in build_genes(options, genotype, beds):
            print(">%s" % (name,))

            sequence = ""
            for fragment in fragments:
                sequence = flush_fasta(sequence + fragment)

            if sequence:
                print(sequence)

    return 0


###############################################################################
###############################################################################


def read_intervals(filename):
    with open(filename) as bed_file:
        intervals = text.parse_lines_by_contig(bed_file, BEDRecord)

        for (key, beds) in intervals.items():
            bed_tuples = []
            for bed in beds:
                if len(bed) < 6:
                    sys.stderr.write(
                        (
                            "ERROR: Invalid BED record '%r', must "
                            "have at least 6 fields\n"
                        )
                        % (bed,)
                    )
                    return None

                bed_tuples.append(bed)
            intervals[key] = bed_tuples

    return intervals


def parse_intervals(genotype):
    records = {}
    for line in genotype.header:
        match = _VCF_DICT.match(line)
        if not match:
            continue

        key, values = match.groups()
        if key == "contig":
            values = dict(pair.split("=", 1) for pair in values.split(","))
            record = BEDRecord()
            record.contig = values["ID"]
            record.start = 0
            record.end = int(values["length"])
            record.name = record.contig
            record.strand = "+"

            records[record.contig] = [record]

    if not records:
        sys.stderr.write(
            "ERROR: List of contigs not found in VCF header; "
            "specifying --intervals is required!\n"
        )
        return None

    return records


def check_nth_sample(options, genotype):
    parser = pysam.asVCF()

    for contig in genotype.contigs:
        for record in genotype.fetch(contig, parser=parser):
            if len(record) <= options.nth_sample:
                sys.stderr.write(
                    "ERROR: Sample %i selected with --nth-sample,"
                    " but file only contains %i sample(s)!\n"
                    % (options.nth_sample + 1, len(record))
                )
                return False
            return True
    return True


def main(argv):
    prog = "paleomix vcf_to_fasta"
    usage = "%s [options] --genotype in.vcf --intervals in.bed" % (prog,)

    parser = argparse.ArgumentParser(prog=prog, usage=usage)
    parser.add_argument(
        "--genotype",
        required=True,
        metavar="VCF",
        help="Tabix indexed VCF file; by default the first "
        "sample is used in multi-sample VCFs. Use "
        "--nth-sample option to select another sample.",
    )
    parser.add_argument(
        "--nth-sample",
        default=1,
        type=int,
        metavar="NTH",
        help="Use Nth sample from the VCF, with the first sample numbered 1",
    )
    parser.add_argument(
        "--intervals",
        metavar="BED",
        help="Six column BED file; sequences on the same "
        "contig with the same name are assumed to "
        "represent the same gene, and are merged into a "
        "single contiguous FASTA sequence.",
    )
    parser.add_argument(
        "--padding",
        type=int,
        default=10,
        help="Number of bases to expand intervals, when checking for adjacent indels",
    )
    parser.add_argument(
        "--whole-codon-indels-only",
        action="store_true",
        default=False,
        help="If true, only indels where (length %% 3) == 0 are retained",
    )
    parser.add_argument(
        "--ignore-indels",
        action="store_true",
        default=False,
        help="Do not include indels generated FASTA sequence.",
    )

    opts = parser.parse_args(argv)

    print("Running vcf_to_fasta", end="", file=sys.stderr)
    if opts.whole_codon_indels_only:
        print(", assuming sequences represents CDS", end="", file=sys.stderr)
    print(file=sys.stderr)

    if not os.path.exists(opts.genotype):
        sys.stderr.write("ERROR: VCF file does not exist.\n")
        return 1
    elif not os.path.exists(opts.genotype + ".tbi"):
        sys.stderr.write("ERROR: VCF file not tabix indexed.\n")
        sys.stderr.write('       To index, run "tabix -p vcf <filename>".\n')
        return 1
    elif opts.nth_sample < 1:
        sys.stderr.write("ERROR: --nth-sample uses 1-based offsets, zero and\n")
        sys.stderr.write("       negative values are not allowed!\n")
        return 1

    # Relevant VCF functions uses zero-based offsets
    opts.nth_sample -= 1

    genotype = pysam.TabixFile(opts.genotype)

    if opts.intervals is None:
        intervals = parse_intervals(genotype)
    else:
        intervals = read_intervals(opts.intervals)

    if intervals is None:
        return 1

    if not check_nth_sample(opts, genotype):
        return 1

    return genotype_genes(opts, intervals, genotype)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from unittest.mock import Mock

import pytest

import paleomix.common.sampling as sampling


###############################################################################
###############################################################################
# weighted_sampling

_SELECT_BY_VALUE = (
    (0.00000, "a"),
    (0.16666, "a"),  # < 1/6
    (1 / 6.0, "b"),
    (0.49999, "b"),  # < 3/6
    (3 / 6.0, "c"),
    (0.99999, "c"),
)


@pytest.mark.parametrize("value, expectation", _SELECT_BY_VALUE)
def test_weighted_sampling__select_by_weight(value, expectation):
    choices = "abc"
    weights = (1, 2, 3)
    rng = Mock(random=lambda: value)
    iterator = sampling.weighted_sampling(choices, weights, rng)
    assert next(iterator) == expectation


@pytest.mark.parametrize("choices, weights", (([], []), ([], [1, 2]), ([1, 2], [])))
def test_weighted_sampling__(choices, weights):
    iterator = sampling.weighted_sampling(choices, weights)
    with pytest.raises(ValueError):
        iterator.__next__()


_MISMATCHED_LENGTH_INPUTS = (
    ([0, 1], [1, 2, 3]),
    ([0, 1, 2], [1, 2]),
    (iter([0, 1]), [1, 2, 3]),
    ([0, 1], iter([1, 2, 3])),
    (iter([0, 1]), iter([1, 2, 3])),
)


@pytest.mark.parametrize("choices, weights", _MISMATCHED_LENGTH_INPUTS)
def test_weighted_sampling__different_length_input_raises_value_error(choices, weights):
    iterator = sampling.weighted_sampling(choices, weights)
    with pytest.raises(ValueError):
        iterator.__next__()


def test_weighted_sampling__negative_weight_value_error():
    choices = [0, 1, 2]
    weights = [1, -2, 3]
    iterator = sampling.weighted_sampling(choices, weights)
    with pytest.raises(ValueError):
        iterator.__next__()


def test_weighted_sampling__zero_weight_raises_value_error():
    choices = [0, 1, 2]
    weights = [1, 0, 3]
    iterator = sampling.weighted_sampling(choices, weights)
    with pytest.raises(ValueError):
        iterator.__next__()


def test_weighted_sampling__non_numerical_weight_raises_type_error():
    choices = [0, 1, 2]
    weights = [1, "foo", 3]
    iterator = sampling.weighted_sampling(choices, weights)
    with pytest.raises(TypeError):
        iterator.__next__()


###############################################################################
###############################################################################
# reservoir_sampling


def test_reservoir_sampling__select_first_item():
    rng = Mock(randint=lambda _min, _max: 1)
    values = [1, 2]
    result = sampling.reservoir_sampling(values, 1, rng)
    assert result == [1]


def test_reservoir_sampling__select_second_item():
    rng = Mock(randint=lambda _min, _max: 0)
    values = [1, 2]
    result = sampling.reservoir_sampling(values, 1, rng)
    assert result == [2]


def test_reservoir_sampling__upsample_equals_input():
    result = sampling.reservoir_sampling(list(range(5)), 10)
    assert result == list(range(5))


def test_reservoir_sampling__downsample_to_zero():
    result = sampling.reservoir_sampling(list(range(5)), 0)
    assert result == []


def test_reservoir_sampling__downsample_to_negative_raises_value_error():
    with pytest.raises(ValueError):
        sampling.reservoir_sampling(list(range(5)), -1)


def test_reservoir_sampling__downsample_to_float_raises_type_error():
    with pytest.raises(TypeError):
        sampling.reservoir_sampling(list(range(5)), 1.0)


def test_reservoir_sampling__downsample_to_non_number_raises_type_error():
    with pytest.raises(TypeError):
        sampling.reservoir_sampling(list(range(5)), "Eh?")



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.common.formats.newick import Newick
from paleomix.common.utilities import safe_coerce_to_tuple
from paleomix.common.fileutils import describe_files, move_file
from paleomix.node import Node


class NewickRerootNode(Node):
    def __init__(self, tree_files, output_file, taxa=(), dependencies=()):
        self._output_file = output_file
        self._tree_files = safe_coerce_to_tuple(tree_files)
        self._reroot_on_taxa = safe_coerce_to_tuple(taxa)

        reroot_on = "midpoint"
        if self._reroot_on_taxa:
            reroot_on = repr("', '".join(sorted(self._reroot_on_taxa)))

        Node.__init__(
            self,
            description="rerooting newick tree %s on %s"
            % (describe_files(tree_files), reroot_on),
            input_files=self._tree_files,
            output_files=self._output_file,
            dependencies=dependencies,
        )

    def _run(self, _config, temp):
        lines = []
        for tree in _read_tree_files(self._tree_files):
            if self._reroot_on_taxa:
                rooted_tree = tree.reroot_on_taxa(self._reroot_on_taxa)
            else:
                rooted_tree = tree.reroot_on_midpoint()
            lines.append(str(rooted_tree))
        lines = "\n".join(lines) + "\n"

        temp_output_file = os.path.join(temp, os.path.basename(self._output_file))
        with open(temp_output_file, "w") as handle:
            handle.write(lines)

        move_file(temp_output_file, self._output_file)


class NewickSupportNode(Node):
    def __init__(
        self, main_tree_files, support_tree_files, output_file, dependencies=()
    ):
        self._output_file = output_file
        self._main_tree_files = safe_coerce_to_tuple(main_tree_files)
        self._support_tree_files = safe_coerce_to_tuple(support_tree_files)
        input_files = self._main_tree_files + self._support_tree_files

        Node.__init__(
            self,
            description="calculating newick support for %s"
            % (describe_files(main_tree_files),),
            input_files=input_files,
            output_files=output_file,
            dependencies=dependencies,
        )

    def _run(self, _config, temp):
        main_trees = _read_tree_files(self._main_tree_files)
        support_trees = _read_tree_files(self._support_tree_files)

        lines = []
        for main_tree in main_trees:
            supported_tree = main_tree.add_support(support_trees)
            lines.append(str(supported_tree))
        lines = "\n".join(lines) + "\n"

        temp_output_file = os.path.join(temp, os.path.basename(self._output_file))
        with open(temp_output_file, "w") as handle:
            handle.write(lines)

        move_file(temp_output_file, self._output_file)


def _read_tree_files(filenames):
    trees = []
    for filename in filenames:
        with open(filename) as handle:
            for line in handle:
                trees.append(Newick.from_string(line))
    return trees



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import operator

import pytest


import paleomix.common.utilities as utils


################################################################################
################################################################################
# Tests for 'safe_coerce_to_tuple'


def test_safe_coerce_to_tuple__str():
    assert utils.safe_coerce_to_tuple("foo") == ("foo",)


def test_safe_coerce_to_tuple__int():
    assert utils.safe_coerce_to_tuple(17) == (17,)


def test_safe_coerce_to_tuple__list():
    assert utils.safe_coerce_to_tuple([1, 3, 2]) == (1, 3, 2)


def test_safe_coerce_to_tuple__tuple():
    assert utils.safe_coerce_to_tuple((1, 3, 2)) == (1, 3, 2)


def test_safe_coerce_to_tuple__iterable():
    assert utils.safe_coerce_to_tuple(iter(range(3))) == (0, 1, 2)


def test_safe_coerce_to_tuple__dict():
    assert utils.safe_coerce_to_tuple({1: 2, 3: 4}) == ({1: 2, 3: 4},)


###############################################################################
###############################################################################
# Tests for 'safe_coerce_to_frozenset'


def test_safe_coerce_to_frozenset__str():
    assert utils.safe_coerce_to_frozenset("foo") == frozenset(("foo",))


def test_safe_coerce_to_frozenset__unicode():
    assert utils.safe_coerce_to_frozenset("foo") == frozenset(("foo",))


def test_safe_coerce_to_frozenset__int():
    assert utils.safe_coerce_to_frozenset(17) == frozenset((17,))


def test_safe_coerce_to_frozenset__list():
    assert utils.safe_coerce_to_frozenset([1, 3, 2]) == frozenset((1, 3, 2))


def test_safe_coerce_to_frozenset__tuple():
    assert utils.safe_coerce_to_frozenset((1, 3, 2)) == frozenset(((1, 3, 2)))


def test_safe_coerce_to_frozenset__iterable():
    assert utils.safe_coerce_to_frozenset(range(3)) == frozenset((0, 1, 2))


def test_safe_coerce_to_frozenset__dict():
    with pytest.raises(TypeError):
        utils.safe_coerce_to_frozenset({1: 2, 3: 4})


###############################################################################
###############################################################################
# Tests for 'try_cast'


def test_try_cast__int_to_int():
    assert utils.try_cast(17, int) == 17


def test_try_cast__float_to_int():
    assert utils.try_cast(17.3, int) == 17


def test_try_cast__good_str_to_int():
    assert utils.try_cast("17", int) == 17


def test_try_cast__bad_str_to_int():
    assert utils.try_cast("x17", int) == "x17"


def test_try_cast__list_to_int():
    assert utils.try_cast([1, 2, 3], int) == [1, 2, 3]


###############################################################################
###############################################################################
# Tests for 'set_in'


def test_set_in__single_kw_in_empty_dictionary():
    value = {}
    utils.set_in(value, ["Foo"], 17)
    assert value == {"Foo": 17}


def test_set_in__two_kws_in_empty_dictionary():
    value = {}
    utils.set_in(value, ["Foo", 13], 17)
    assert value == {"Foo": {13: 17}}


def test_set_in__three_kws_in_empty_dictionary():
    value = {}
    utils.set_in(value, ["Foo", 13, (1, 2)], 17)
    assert value == {"Foo": {13: {(1, 2): 17}}}


def test_set_in__three_kws_in_partial_dictionary():
    value = {"Foo": {12: 0}}
    utils.set_in(value, ["Foo", 13, (1, 2)], 17)
    assert value == {"Foo": {12: 0, 13: {(1, 2): 17}}}

    value = {"Foo": {13: {"Bar": None}}}
    utils.set_in(value, ["Foo", 13, (1, 2)], 17)
    assert value == {"Foo": {13: {(1, 2): 17, "Bar": None}}}


def test_set_in__update_value_one_kw():
    value = {1: None}
    utils.set_in(value, [1], 3.14)
    assert value == {1: 3.14}


def test_set_in__update_value_two_kw():
    value = {1: {2: 3}}
    utils.set_in(value, [1, 2], 365)
    assert value == {1: {2: 365}}


def test_set_in__fail_on_no_kws():
    with pytest.raises(ValueError):
        utils.set_in({}, [], 17)


def test_set_in__fail_on_invalid_sub_dictionary_first_level():
    with pytest.raises(TypeError):
        utils.set_in(None, [1], 17)


def test_set_in__fail_on_invalid_sub_dictionary_second_level():
    with pytest.raises(TypeError):
        utils.set_in({1: None}, [1, 2], 17)


def test_set_in__fail_on_invalid_sub_dictionary_third_level():
    with pytest.raises(TypeError):
        utils.set_in({1: {2: None}}, [1, 2, 3], 17)


def test_set_in__iteratable_keywords():
    value = {}
    utils.set_in(value, iter(["Foo", 13, (1, 2)]), 17)
    assert value == {"Foo": {13: {(1, 2): 17}}}


###############################################################################
###############################################################################
# Tests for 'get_in'


def test_get_in__get_value_one_keyword():
    assert utils.get_in({1: 2}, [1]) == 2


def test_get_in__get_value_two_keywords():
    assert utils.get_in({1: {2: 3}}, [1, 2]) == 3


def test_get_in__get_value_three_keywords():
    assert utils.get_in({1: {2: {3: 4}}}, [1, 2, 3]) == 4


def test_get_in__get_default_one_keyword():
    assert utils.get_in({1: 2}, [2]) is None


def test_get_in__get_default_one_keyword_with_default():
    assert utils.get_in({1: 2}, [2], "other") == "other"


def test_get_in__get_default_three_keywords_fail_at_first():
    assert utils.get_in({1: {2: {3: 4}}}, [2, 2, 4]) is None


def test_get_in__get_default_three_keywords_fail_at_first_with_default():
    assert utils.get_in({1: {2: {3: 4}}}, [2, 2, 4], "other") == "other"


def test_get_in__get_default_three_keywords_fail_at_second():
    assert utils.get_in({1: {2: {3: 4}}}, [1, 3, 4]) is None


def test_get_in__get_default_three_keywords_fail_at_second_with_default():
    assert utils.get_in({1: {2: {3: 4}}}, [1, 3, 4], "other") == "other"


def test_get_in__get_default_three_keywords_fail_at_third():
    assert utils.get_in({1: {2: {3: 4}}}, [1, 2, 4]) is None


def test_get_in__get_default_three_keywords_fail_at_third_with_default():
    assert utils.get_in({1: {2: {3: 4}}}, [1, 2, 4], "other") == "other"


def test_get_in__iterator_keywords():
    assert utils.get_in({1: {2: {3: 4}}}, iter([1, 2, 3])) == 4


###############################################################################
###############################################################################
# Tests for 'split_before'


def _do_split(lst, key):
    # Convertion to list allows the implementation to be
    # lazy, while making comparisons for asserts easier
    return list(utils.split_before(lst, key))


def test_split_before__split_empty_list():
    assert _do_split([], None) == []


def test_split_before__split_list_with_no_true_pred():
    assert _do_split(list(range(10)), lambda x: False) == [list(range(10))]


def test_split_before__split_list_true_pred_at_first_position():
    assert _do_split(list(range(4)), lambda x: x % 2 == 0) == [[0, 1], [2, 3]]


def test_split_before__split_list_true_pred_at_second_position():
    assert _do_split(list(range(4)), lambda x: x % 2 == 1) == [[0], [1, 2], [3]]


def test_split_before__split_consequtive_true_pred():
    assert _do_split(list(range(0, 5, 2)), lambda x: x % 2 == 0) == [[0], [2], [4]]


def test_split_before__no_hits():
    assert _do_split(list(range(1, 5)), lambda x: x % 5 == 0) == [list(range(1, 5))]


###############################################################################
###############################################################################
# Tests for 'grouper'


def test_grouper__empty_list():
    result = utils.grouper(3, [])
    assert list(result) == []


def test_grouper__non_empty_list():
    result = utils.grouper(3, list(range(6)))
    expected = [(0, 1, 2), (3, 4, 5)]
    assert list(result) == expected


def test_grouper__non_empty_list_with_trailing():
    result = utils.grouper(3, list(range(7)))
    expected = [(0, 1, 2), (3, 4, 5), (6, None, None)]
    assert list(result) == expected


def test_grouper__non_empty_list_with_trailing_fill_value():
    result = utils.grouper(3, list(range(7)), fillvalue=r"\0")
    expected = [(0, 1, 2), (3, 4, 5), (6, r"\0", r"\0")]
    assert list(result) == expected


###############################################################################
###############################################################################
# Tests for 'group_by_pred'


def test_group_by_pred__empty_list():
    assert utils.group_by_pred(id, []) == ([], [])


def test_group_by_pred__always_false():
    assert utils.group_by_pred(lambda x: False, [1, 2, 3]) == ([], [1, 2, 3])


def test_group_by_pred__always_true():
    assert utils.group_by_pred(lambda x: True, [1, 2, 3]) == ([1, 2, 3], [])


def test_group_by_pred__is_even():
    assert utils.group_by_pred(lambda x: x % 2 == 0, [1, 2, 3]) == ([2], [1, 3])


def test_group_by_pred__iterable():
    assert utils.group_by_pred(lambda x: x % 2 == 0, range(1, 4)) == ([2], [1, 3])


###############################################################################
###############################################################################
# Tests for 'fragment'


def test_fragment__empty():
    assert list(utils.fragment(5, "")) == []
    assert list(utils.fragment(5, [])) == []


def test_fragment__partial_fragment():
    assert list(utils.fragment(3, "ab")) == ["ab"]
    assert list(utils.fragment(3, ["a", "b"])) == [["a", "b"]]


def test_fragment__single_fragment():
    assert list(utils.fragment(3, "abc")) == ["abc"]
    assert list(utils.fragment(3, ["a", "b", "c"])) == [["a", "b", "c"]]


def test_fragment__multiple_fragments():
    assert list(utils.fragment(3, "abcdef")) == ["abc", "def"]
    assert list(utils.fragment(3, list("abcdef"))) == [list("abc"), list("def")]


def test_fragment__multiple_fragments_partial():
    assert list(utils.fragment(3, "abcdefgh")) == ["abc", "def", "gh"]
    assert list(utils.fragment(3, list("abcdefgh"))) == [
        list("abc"),
        list("def"),
        list("gh"),
    ]


def test_fragment__range():
    assert list(utils.fragment(3, range(6))) == [range(3), range(3, 6)]


def test_fragment__iterable():
    with pytest.raises(TypeError):
        list(utils.fragment(3, iter(range(6))))


def test_fragment__set():
    with pytest.raises(TypeError):
        list(utils.fragment(3, set(range(6))))


###############################################################################
###############################################################################
# Tests for 'cumsum'


def test_cumsum__empty():
    assert list(utils.cumsum([])) == []


def test_cumsum__integers():
    assert list(utils.cumsum(list(range(-4, 5)))) == [
        -4,
        -7,
        -9,
        -10,
        -10,
        -9,
        -7,
        -4,
        0,
    ]


def test_cumsum__float():
    assert list(utils.cumsum((1.0, 2.0, 3.0))) == [1.0, 3.0, 6.0]


def test_cumsum__initial():
    assert list(utils.cumsum(list(range(5)), -10)) == [-10, -9, -7, -4, 0]


###############################################################################
###############################################################################
# fill_dict


def test_fill_dict__empty_dicts():
    result = utils.fill_dict({}, {})
    assert result == {}


def test_fill_dict__filling_empty_dict():
    source = {"a": 1, "b": {"c": 2, "d": 3}}
    expected = {"a": 1, "b": {"c": 2, "d": 3}}
    result = utils.fill_dict({}, source)
    assert result == expected


def test_fill_dict__filling_full_dict():
    source = {"a": 1, "b": {"c": 2, "d": 3}}
    destination = {"a": 2, "b": {"c": 3, "d": 4}}
    expected = {"a": 2, "b": {"c": 3, "d": 4}}
    result = utils.fill_dict(destination, source)
    assert result == expected


def test_fill_dict__destination_not_modified():
    source = {"a": 1, "b": {"c": 2, "d": 3}}
    destination = {"b": {"d": 0}}
    utils.fill_dict(destination, source)
    assert destination == {"b": {"d": 0}}


def test_fill_dict__source_not_modified():
    expected = {"a": 1, "b": {"c": 2, "d": 3}}
    source = {"a": 1, "b": {"c": 2, "d": 3}}
    destination = {"b": {"d": 0}}
    utils.fill_dict(destination, source)
    assert source == expected


def test_fill_dict__destination_must_be_dict():
    with pytest.raises(TypeError):
        utils.fill_dict([], {})


def test_fill_dict__source_must_be_dict():
    with pytest.raises(TypeError):
        utils.fill_dict({}, [])


###############################################################################
###############################################################################
# chain_sorted


def test_chain_sorted__no_sequences():
    expected = ()
    result = tuple(utils.chain_sorted())
    assert expected == result


def test_chain_sorted__single_sequence():
    sequence = (1, 2, 3)
    result = tuple(utils.chain_sorted(sequence))
    assert sequence == result


_SEQUENTIAL_CONTENT_1 = (1, 2, 3)
_SEQUENTIAL_CONTENT_2 = (4, 5, 6)
_SEQUENTIAL_CONTENT_PERMUTATIONS = (
    (_SEQUENTIAL_CONTENT_1, _SEQUENTIAL_CONTENT_2),
    (_SEQUENTIAL_CONTENT_2, _SEQUENTIAL_CONTENT_1),
)


@pytest.mark.parametrize("seq_a, seq_b", _SEQUENTIAL_CONTENT_PERMUTATIONS)
def test_chain_sorted__sequential_contents(seq_a, seq_b):
    expected = (1, 2, 3, 4, 5, 6)
    result = tuple(utils.chain_sorted(seq_a, seq_b))
    assert expected == result


def test_chain_sorted__mixed_contents():
    sequence_a = (3, 4, 8)
    sequence_c = (0, 1, 6)
    sequence_b = (2, 5, 7)
    expected = (0, 1, 2, 3, 4, 5, 6, 7, 8)
    result = tuple(utils.chain_sorted(sequence_a, sequence_b, sequence_c))
    assert expected == result


def test_chain_sorted__mixed_length_contents():
    sequence_a = (1,)
    sequence_c = (0, 2)
    sequence_b = ()
    expected = (0, 1, 2)
    result = tuple(utils.chain_sorted(sequence_a, sequence_b, sequence_c))
    assert expected == result


def test_chain_sorted__mixed_contents__key():
    sequence_a = (-2, -3, -5)
    sequence_b = (0, -1, -4)
    expected = (0, -1, -2, -3, -4, -5)
    result = tuple(utils.chain_sorted(sequence_a, sequence_b, key=abs))
    assert expected == result


def test_chain_sorted__identical_objects_are_preserved():
    object_a = [1]
    object_b = [1]
    assert object_a is not object_b
    expected = (object_a, object_b)
    result = tuple(utils.chain_sorted([object_a], [object_b]))
    assert expected == result
    assert object_a is result[0] or object_a is result[1]
    assert object_b is result[0] or object_b is result[1]


def test_chain_sorted__stable_sort():
    object_a = [1]
    object_b = [1]
    object_c = [2]
    object_d = [2]
    seq_a = [object_a, object_c]
    seq_b = [object_b, object_d]

    expected = (object_a, object_b, object_c, object_d)
    result = tuple(utils.chain_sorted(seq_a, seq_b))
    assert expected == result
    assert all(a is b for (a, b) in zip(expected, result))

    expected = (object_b, object_a, object_d, object_c)
    result = tuple(utils.chain_sorted(seq_b, seq_a))
    assert expected == result
    assert all(a is b for (a, b) in zip(expected, result))


def test_chain_sorted__runs_of_values():
    object_a = [1]
    object_b = [1]
    object_c = [2]
    object_d = [2]
    seq_a = [object_a, object_b]
    seq_b = [object_c, object_d]

    expected = (object_a, object_b, object_c, object_d)
    result = tuple(utils.chain_sorted(seq_a, seq_b))
    assert expected == result
    assert all(a is b for (a, b) in zip(expected, result))


def test_chain_sorted__invalid_keywords():
    with pytest.raises(TypeError):
        tuple(utils.chain_sorted((1, 2, 3), foobar=None))


###############################################################################
###############################################################################
# Immutable


def test_immutable__properties_set():
    class ImmutableCls(utils.Immutable):
        def __init__(self, value):
            utils.Immutable.__init__(self, value=value)

    obj = ImmutableCls(17)
    assert obj.value == 17


@pytest.mark.parametrize("key, value", (("value", 13), ("new_value", "foo")))
def test_immutable__properties_immutable(key, value):
    class ImmutableCls(utils.Immutable):
        def __init__(self, value):
            utils.Immutable.__init__(self, value=value)

    obj = ImmutableCls(17)
    with pytest.raises(NotImplementedError):
        setattr(obj, key, value)


def test_immutable__properties_del():
    class ImmutableCls(utils.Immutable):
        def __init__(self, value):
            utils.Immutable.__init__(self, value=value)

    def _del_property(obj):
        del obj.value

    obj = ImmutableCls(17)
    with pytest.raises(NotImplementedError):
        _del_property(obj)


###############################################################################
###############################################################################
# TotallyOrdered


class SomethingOrdered(utils.TotallyOrdered):
    def __init__(self, value):
        self.value = value

    def __lt__(self, other):
        if not isinstance(other, SomethingOrdered):
            return NotImplemented

        return self.value < other.value


# Less than
def test_totally_ordered__lt_vs_lt():
    assert SomethingOrdered(1) < SomethingOrdered(2)


def test_totally_ordered__lt_vs_gt():
    assert not (SomethingOrdered(1) < SomethingOrdered(0))


def test_totally_ordered__lt_vs_wrong_type():
    assert SomethingOrdered(1).__lt__("Foo") == NotImplemented


# Less than or equal
def test_totally_ordered__le_vs_le():
    assert SomethingOrdered(1) <= SomethingOrdered(1)


def test_totally_ordered__le_vs_gt():
    assert not (SomethingOrdered(1) <= SomethingOrdered(0))


def test_totally_ordered__le_vs_wrong_type():
    assert SomethingOrdered(1).__le__("Foo") == NotImplemented


# Greater than or equal
def test_totally_ordered__ge_vs_ge():
    assert SomethingOrdered(1) >= SomethingOrdered(1)


def test_totally_ordered__ge_vs_lt():
    assert not (SomethingOrdered(0) >= SomethingOrdered(1))


def test_totally_ordered__ge_vs_wrong_type():
    assert SomethingOrdered(1).__ge__("Foo") == NotImplemented


# Greater than
def test_totally_ordered__gt_vs_gt():
    assert SomethingOrdered(1) > SomethingOrdered(0)


def test_totally_ordered__gt_vs_eq():
    assert not (SomethingOrdered(0) > SomethingOrdered(0))


def test_totally_ordered__gt_vs_wrong_type():
    assert SomethingOrdered(1).__gt__("Foo") == NotImplemented


# Equal to
def test_totally_ordered__eq_vs_eq():
    assert SomethingOrdered(1) == SomethingOrdered(1)


def test_totally_ordered__eq_vs_ne():
    assert not (SomethingOrdered(1) == SomethingOrdered(2))


def test_totally_ordered__eq_vs_wrong_type():
    assert SomethingOrdered(1).__eq__("Foo") == NotImplemented


# Not equal to
def test_totally_ordered__ne_vs_ne():
    assert SomethingOrdered(1) != SomethingOrdered(2)


def test_totally_ordered__ne_vs_eq():
    assert not (SomethingOrdered(1) != SomethingOrdered(1))


def test_totally_ordered__ne_vs_wrong_type():
    assert SomethingOrdered(1).__ne__("Foo") == NotImplemented


class SomethingBadlyOrdered(utils.TotallyOrdered):
    def __init__(self, value):
        self.value = value


def test_totally_ordered__missing_implementation():
    obj_a = SomethingBadlyOrdered(1)
    obj_b = SomethingBadlyOrdered(2)
    with pytest.raises(NotImplementedError):
        operator.gt(obj_a, obj_b)



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy

import pytest

from paleomix.common.bedtools import (
    BEDError,
    BEDRecord,
    merge_bed_records,
    pad_bed_records,
)

###############################################################################
###############################################################################
# BEDRecord constructor


def test_bedrecord__constructor__defaults():
    record = BEDRecord()
    assert len(record) == 0
    assert str(record) == ""
    assert repr(record) == "BEDRecord()"


def test_bedrecord__constructor__empty_string():
    record = BEDRecord("")
    assert len(record) == 0
    assert str(record) == ""
    assert repr(record) == "BEDRecord()"


def test_bedrecord__constructor__3_fields():
    text = "my_contig\t12\t345"
    record = BEDRecord(text)

    assert len(record) == 3
    assert str(record) == text
    assert repr(record) == "BEDRecord(contig='my_contig', start=12, end=345)"


def test_bedrecord__constructor__6_fields():
    text = "my_contig\t12\t345\tmy_name\t-3\t-"
    record = BEDRecord(text)

    assert len(record) == 6
    assert str(record) == text
    assert (
        repr(record) == "BEDRecord(contig='my_contig', start=12, "
        "end=345, name='my_name', score=-3, strand='-')"
    )


def test_bedrecord__constructor__extra_fields():
    text = "my_contig\t12\t345\tmy_name\t-3\t-\tfoo\tbar"
    record = BEDRecord(text)

    assert len(record) == 8
    assert str(record) == text
    assert (
        repr(record) == "BEDRecord(contig='my_contig', start=12, "
        "end=345, name='my_name', score=-3, strand='-', "
        "'foo', 'bar')"
    )


def test_bedrecord__constructor__invalid_values():
    tmpl = "\t1\t%s\t\t0\t-"
    BEDRecord(tmpl % (1,))  # check template

    with pytest.raises(BEDError):
        BEDRecord(tmpl % ("not a number",))


###############################################################################
###############################################################################
# BEDRecord accessors


def test_bedrecord__accessors__3_fields():
    record = BEDRecord("my_contig\t12\t345")

    assert record.contig == "my_contig"
    assert record.start == 12
    assert record.end == 345
    with pytest.raises(IndexError):
        record.name()
    with pytest.raises(IndexError):
        record.score()
    with pytest.raises(IndexError):
        record.strand()


def test_bedrecord__accessors__6_fields():
    record = BEDRecord("my_contig\t12\t345\tmy_name\t-3\t-")

    assert record.contig == "my_contig"
    assert record.start == 12
    assert record.end == 345
    assert record.name == "my_name"
    assert record.score == -3
    assert record.strand == "-"


def test_bedrecord__accessors__extra_fields():
    text = "my_contig\t12\t345\tmy_name\t-3\t-\tfoo\tbar"
    record = BEDRecord(text)

    assert record[6] == "foo"
    assert record[7] == "bar"


def test_bedrecord__accessors__6_fields__getitem():
    record = BEDRecord("my_contig\t12\t345\tmy_name\t-3\t-")

    assert record[0] == "my_contig"
    assert record[1] == 12
    assert record[2] == 345
    assert record[3] == "my_name"
    assert record[4] == -3
    assert record[5] == "-"


def test_bedrecord__setters__3_fields():
    record = BEDRecord("my_contig\t12\t345")

    record.contig = "chrZ"
    assert record.contig == "chrZ"

    record.end += 20
    assert record.end == 365

    assert str(record) == "chrZ\t12\t365"
    assert repr(record) == "BEDRecord(contig='chrZ', start=12, end=365)"


def test_bedrecord__setters__type_errors():
    record = BEDRecord("my_contig\t12\t345\tname\t0\t+")

    with pytest.raises(ValueError):
        record.contig = 17
    with pytest.raises(ValueError):
        record.start = "foo"
    with pytest.raises(ValueError):
        record.end = "foo"
    with pytest.raises(ValueError):
        record.name = 17.3
    with pytest.raises(ValueError):
        record.score = "foo"
    with pytest.raises(ValueError):
        record.strand = "foo"


def test_bedrecord__setters__unset_fields__at_end():
    record = BEDRecord("my_contig\t12\t345")

    record.name = "my_region"
    assert record.name == "my_region"

    record.score = -13
    assert record.score == -13

    record.strand = "-"
    assert record.strand == "-"

    assert str(record) == "my_contig\t12\t345\tmy_region\t-13\t-"
    assert (
        repr(record) == "BEDRecord(contig='my_contig', start=12, end=345, "
        "name='my_region', score=-13, strand='-')"
    )


def test_bedrecord__setters__unset_fields__after_end():
    record = BEDRecord("")
    record.strand = "-"
    assert str(record) == "\t0\t0\t\t0\t-"

    record = BEDRecord("my_name")
    record.strand = "-"
    assert str(record) == "my_name\t0\t0\t\t0\t-"

    record = BEDRecord("my_name\t17")
    record.strand = "-"
    assert str(record) == "my_name\t17\t0\t\t0\t-"

    record = BEDRecord("my_name\t17\t258")
    record.strand = "-"
    assert str(record) == "my_name\t17\t258\t\t0\t-"

    record = BEDRecord("my_name\t17\t258\tregion")
    record.strand = "-"
    assert str(record) == "my_name\t17\t258\tregion\t0\t-"

    record = BEDRecord("my_name\t17\t258\tregion\t33")
    record.strand = "-"
    assert str(record) == "my_name\t17\t258\tregion\t33\t-"

    record = BEDRecord("my_name\t17\t258\tregion\t33\t+")
    record.strand = "-"
    assert str(record) == "my_name\t17\t258\tregion\t33\t-"


def test_bedrecord__cmp():
    record_1_txt = "my_contig\t12\t345\tmy_name\t-3\t-\tfoo"
    record_1 = BEDRecord(record_1_txt)
    record_2 = BEDRecord("chrZ\t132\t4345\tchrZ_region\t0\t+\tbar")

    for idx in range(len(record_2)):
        record_tmp = BEDRecord(record_1_txt)
        assert record_1 == record_tmp
        record_tmp[idx] = record_2[idx]
        assert record_1 != record_tmp
        record_tmp[idx] = record_1[idx]
        assert record_1 == record_tmp


###############################################################################
# freeze


def test_bedrecord__freeze__modify_existing_value():
    record_str = "my_contig\t12\t345\tmy_name"
    record = BEDRecord(record_str)
    frozen_record = record.freeze()

    with pytest.raises(TypeError):
        frozen_record.start = 1

    assert str(frozen_record) == record_str


def test_bedrecord__freeze__add_new_value():
    record_str = "my_contig\t12\t345\tmy_name"
    record = BEDRecord(record_str)
    frozen_record = record.freeze()

    with pytest.raises(AttributeError):
        frozen_record.score = 1

    assert str(frozen_record) == record_str


###############################################################################
###############################################################################


def test_bedrecord__copy():
    record_1_txt = "my_contig\t12\t345\tmy_name\t-3\t-"
    record_1 = BEDRecord(record_1_txt)
    record_2 = copy.copy(record_1)
    record_2.name = "my_clone"

    assert str(record_1) == record_1_txt
    assert str(record_2) == "my_contig\t12\t345\tmy_clone\t-3\t-"


def test_bedrecord__deepcopy():
    record_1_txt = "my_contig\t12\t345\tmy_name\t-3\t-"
    record_1 = BEDRecord(record_1_txt)
    record_1[6] = ["foo"]
    record_2 = copy.deepcopy(record_1)
    record_2[6][0] = "bar"

    assert str(record_1) == record_1_txt + "\t['foo']"
    assert str(record_2) == record_1_txt + "\t['bar']"


###############################################################################
# pad_bed_records


def test_pad_bed_records__empty_sequences():
    pad_bed_records([], 10)
    pad_bed_records((), 10)


def test_pad_bed_records():
    records = [
        _new_bed_record("chr1", 10, 90),
        _new_bed_record("chr2", 100, 200),
    ]

    pad_bed_records(records, 20)
    assert records == [
        _new_bed_record("chr1", 0, 110),
        _new_bed_record("chr2", 80, 220),
    ]


def test_pad_bed_records__with_max_lengths():
    max_sizes = {"chr1": 100, "chr2": 200}
    records = [
        _new_bed_record("chr1", 10, 90),
        _new_bed_record("chr2", 10, 90),
        _new_bed_record("chr2", 100, 190),
    ]

    pad_bed_records(records, 20, max_sizes)
    assert records == [
        _new_bed_record("chr1", 0, 100),
        _new_bed_record("chr2", 0, 110),
        _new_bed_record("chr2", 80, 200),
    ]


###############################################################################
# merge_bed_records


def test_merge_records__empty_sequences():
    assert merge_bed_records(()) == []
    assert merge_bed_records([]) == []


def test_merge_records__single_record():
    assert merge_bed_records([_new_bed_record("chr1", 1234, 5678)]) == [
        _new_bed_record("chr1", 1234, 5678)
    ]


def test_merge_records__minimal_fields_only():
    assert merge_bed_records([_new_bed_record("chr1", 1234, 5678, "foo", 1, "-")]) == [
        _new_bed_record("chr1", 1234, 5678)
    ]


def test_merge_records__overlapping_records_1():
    assert merge_bed_records(
        [_new_bed_record("chr1", 1234, 5678), _new_bed_record("chr1", 5677, 9012)]
    ) == [_new_bed_record("chr1", 1234, 9012)]


def test_merge_records__overlapping_records_2():
    assert merge_bed_records(
        [_new_bed_record("chr1", 1234, 5678), _new_bed_record("chr1", 5678, 9012)]
    ) == [_new_bed_record("chr1", 1234, 9012)]


def test_merge_records__non_overlapping_records_1():
    assert merge_bed_records(
        [_new_bed_record("chr1", 1234, 5678), _new_bed_record("chr1", 5679, 9012)]
    ) == [_new_bed_record("chr1", 1234, 5678), _new_bed_record("chr1", 5679, 9012)]


def test_merge_records__non_overlapping_records_2():
    assert merge_bed_records(
        [_new_bed_record("chr1", 1234, 5678), _new_bed_record("chr1", 5680, 9012)]
    ) == [_new_bed_record("chr1", 1234, 5678), _new_bed_record("chr1", 5680, 9012)]


def test_merge_records__complex_example():
    assert merge_bed_records(
        [
            _new_bed_record("chr1", 1234, 5678),
            _new_bed_record("chr1", 5678, 9012),
            _new_bed_record("chr2", 1, 20),
            _new_bed_record("chr2", 100, 200),
            _new_bed_record("chr2", 150, 250),
        ]
    ) == [
        _new_bed_record("chr1", 1234, 9012),
        _new_bed_record("chr2", 1, 20),
        _new_bed_record("chr2", 100, 250),
    ]


def test_merge_records__complex_example__unsorted():
    assert merge_bed_records(
        [
            _new_bed_record("chr2", 100, 200),
            _new_bed_record("chr1", 1234, 5678),
            _new_bed_record("chr2", 150, 250),
            _new_bed_record("chr2", 1, 20),
            _new_bed_record("chr1", 5678, 9012),
        ]
    ) == [
        _new_bed_record("chr1", 1234, 9012),
        _new_bed_record("chr2", 1, 20),
        _new_bed_record("chr2", 100, 250),
    ]


def _new_bed_record(*args):
    record = BEDRecord()
    record._fields = list(args)

    return record



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import errno
import sys

import pysam

import paleomix.common.argparse as argparse
import paleomix.common.vcffilter as vcffilter

from paleomix.common.fileutils import open_ro


def _read_files(args):
    in_header = True
    has_filters = False
    vcf_parser = pysam.asVCF()
    for filename in args.filenames:
        with open_ro(filename, "rb") as handle:
            for line in handle:
                if not line.startswith(b"#"):
                    in_header = False
                    line = line.rstrip(b"\n\r")
                    vcf = vcf_parser(line, len(line))
                    if args.reset_filter:
                        vcf.filter = "."

                    yield vcf
                elif in_header:
                    if not (line.startswith(b"##") or has_filters):
                        has_filters = True
                        for item in sorted(vcffilter.describe_filters(args).items()):
                            print('##FILTER=<ID=%s,Description="%s">' % item)

                    print(line.decode("utf-8"), end="")


def main(argv):
    parser = argparse.ArgumentParser(prog="paleomix vcf_filter")

    parser.add_argument(
        "filenames",
        nargs="*",
        help="VCF files; may be gzip/bzip2 compresssed. Leave blank or use '-' to "
        "read from STDIN",
        metavar="file",
    )

    parser.add_argument(
        "--reset-filter",
        default=False,
        action="store_true",
        help="If set, values in the 'FILTER' column  are cleared, and set according "
        "to the results from running the filters implemented by this tool. If not "
        "set, any existing values are retained, and any (new) failed filters are "
        "added to these.",
    )

    vcffilter.add_varfilter_options(parser)
    args = parser.parse_args(argv)

    if (not args.filenames or "-" in args.filenames) and sys.stdin.isatty():
        parser.error("STDIN is a terminal, terminating!")

    try:
        for vcf in vcffilter.filter_vcfs(args, _read_files(args)):
            print(vcf)
    except IOError as error:
        # Check for broken pipe (head, less, etc).
        if error.errno != errno.EPIPE:
            raise


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix.common.rtools as rtools
import paleomix.common.versions as versions

from paleomix.common.fileutils import describe_files

from paleomix.node import NodeError, CommandNode
from paleomix.nodes.samtools import merge_bam_files_command
from paleomix.atomiccmd.builder import AtomicCmdBuilder, apply_options
from paleomix.atomiccmd.sets import ParallelCmds


MAPDAMAGE_VERSION = versions.Requirement(
    call=("mapDamage", "--version"),
    search=r"(\d+)\.(\d+).(\d+)",
    checks=versions.GE(2, 2, 1),
)

RSCRIPT_VERSION = versions.Requirement(
    call=("Rscript", "--version"),
    search=r"version (\d+)\.(\d+).(\d+)",
    checks=versions.GE(3, 3, 3),
    priority=10,
)


class MapDamagePlotNode(CommandNode):
    def __init__(
        self,
        reference,
        input_files,
        output_directory,
        title="mapDamage",
        options={},
        dependencies=(),
    ):
        merge = merge_bam_files_command(input_files)
        command = AtomicCmdBuilder(
            [
                "mapDamage",
                "--no-stats",
                # Prevent references with many contigs from using excessive
                # amounts of memory, at the cost of per-contig statistics:
                "--merge-reference-sequences",
                "-t",
                title,
                "-i",
                "-",
                "-d",
                "%(TEMP_DIR)s",
                "-r",
                "%(IN_REFERENCE)s",
            ],
            IN_STDIN=merge,
            IN_REFERENCE=reference,
            OUT_FREQ_3p=os.path.join(output_directory, "3pGtoA_freq.txt"),
            OUT_FREQ_5p=os.path.join(output_directory, "5pCtoT_freq.txt"),
            OUT_COMP_USER=os.path.join(output_directory, "dnacomp.txt"),
            OUT_PLOT_FRAG=os.path.join(
                output_directory, "Fragmisincorporation_plot.pdf"
            ),
            OUT_PLOT_LEN=os.path.join(output_directory, "Length_plot.pdf"),
            OUT_LENGTH=os.path.join(output_directory, "lgdistribution.txt"),
            OUT_MISINCORP=os.path.join(output_directory, "misincorporation.txt"),
            OUT_LOG=os.path.join(output_directory, "Runtime_log.txt"),
            TEMP_OUT_STDOUT="pipe_mapDamage.stdout",
            TEMP_OUT_STDERR="pipe_mapDamage.stderr",
            CHECK_RSCRIPT=RSCRIPT_VERSION,
            CHECK_MAPDAMAGE=MAPDAMAGE_VERSION,
        )

        apply_options(command, options)

        CommandNode.__init__(
            self,
            command=ParallelCmds([merge, command.finalize()]),
            description="creating mapDamage plots from %s"
            % (describe_files(merge.input_files),),
            dependencies=dependencies,
        )

    def _teardown(self, config, temp):
        # No Length_plot.pdf file is written if there are no SE reads in the
        # input_file. In that case, we write a dummy PDF to ensure that all
        # expected files exist.
        err_message = "No length distributions are available"
        with open(os.path.join(temp, "pipe_mapDamage.stderr")) as in_handle:
            if any(line.startswith(err_message) for line in in_handle):
                fpath = os.path.join(temp, "Length_plot.pdf")
                with open(fpath, "w") as out_handle:
                    out_handle.write(_DUMMY_LENGTH_PLOT_PDF)

        CommandNode._teardown(self, config, temp)


class MapDamageModelNode(CommandNode):
    def __init__(self, reference, directory, options={}, dependencies=()):
        command = AtomicCmdBuilder(
            [
                "mapDamage",
                "--stats-only",
                "-r",
                "%(IN_REFERENCE)s",
                "-d",
                "%(TEMP_DIR)s",
            ],
            IN_REFERENCE=reference,
            TEMP_OUT_FREQ_3p="3pGtoA_freq.txt",
            TEMP_OUT_FREQ_5p="5pCtoT_freq.txt",
            TEMP_OUT_COMP_USER="dnacomp.txt",
            TEMP_OUT_MISINCORP="misincorporation.txt",
            TEMP_OUT_LOG="Runtime_log.txt",
            TEMP_OUT_STDOUT="pipe_mapDamage.stdout",
            TEMP_OUT_STDERR="pipe_mapDamage.stderr",
            OUT_COMP_GENOME=os.path.join(directory, "dnacomp_genome.csv"),
            OUT_MCMC_PROBS=os.path.join(directory, "Stats_out_MCMC_correct_prob.csv"),
            OUT_MCMC_HIST=os.path.join(directory, "Stats_out_MCMC_hist.pdf"),
            OUT_MCMC_ITER=os.path.join(directory, "Stats_out_MCMC_iter.csv"),
            OUT_MCMC_ITERSUM=os.path.join(
                directory, "Stats_out_MCMC_iter_summ_stat.csv"
            ),
            OUT_MCMC_POSTPRED=os.path.join(directory, "Stats_out_MCMC_post_pred.pdf"),
            OUT_MCMC_TRACE=os.path.join(directory, "Stats_out_MCMC_trace.pdf"),
            CHECK_RSCRIPT=RSCRIPT_VERSION,
            CHECK_MAPDAMAGE=MAPDAMAGE_VERSION,
            CHECK_R_INLINE=rtools.requirement("inline"),
            CHECK_R_GGPLOT2=rtools.requirement("ggplot2"),
            CHECK_R_RCPP=rtools.requirement("Rcpp"),
            CHECK_R_GAM=rtools.requirement("gam"),
            CHECK_R_RCPPGSL=rtools.requirement("RcppGSL"),
        )

        apply_options(command, options)

        self._directory = directory

        CommandNode.__init__(
            self,
            command=command.finalize(),
            description="creating mapDamage model from %r" % (directory,),
            dependencies=dependencies,
        )

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)
        for fname in (
            "3pGtoA_freq.txt",
            "5pCtoT_freq.txt",
            "dnacomp.txt",
            "misincorporation.txt",
        ):
            relpath = os.path.join(self._directory, fname)
            abspath = os.path.abspath(relpath)
            os.symlink(abspath, os.path.join(temp, fname))

    def _run(self, config, temp):
        try:
            CommandNode._run(self, config, temp)
        except NodeError as error:
            err_message = "DNA damage levels are too low"
            if self._command.join() == [1]:
                fpath = os.path.join(temp, "pipe_mapDamage.stdout")
                with open(fpath) as handle:
                    for line in handle:
                        if err_message in line:
                            line = line.strip().replace("Warning:", "ERROR:")
                            error = NodeError("%s\n\n%s" % (error, line))
                            break
            raise error


class MapDamageRescaleNode(CommandNode):
    def __init__(
        self,
        reference,
        input_files,
        output_file,
        directory,
        options={},
        dependencies=(),
    ):
        stats_out_fname = "Stats_out_MCMC_correct_prob.csv"

        merge = merge_bam_files_command(input_files)
        command = AtomicCmdBuilder(
            [
                "mapDamage",
                "--rescale-only",
                "-i",
                "-",
                "-d",
                "%(TEMP_DIR)s",
                "-r",
                "%(IN_REFERENCE)s",
                "--rescale-out",
                "%(OUT_BAM)s",
            ],
            IN_STDIN=merge,
            IN_REFERENCE=reference,
            TEMP_OUT_LOG="Runtime_log.txt",
            TEMP_OUT_CSV=stats_out_fname,
            OUT_BAM=output_file,
            CHECK_VERSION=MAPDAMAGE_VERSION,
        )

        apply_options(command, options)

        self._directory = directory

        CommandNode.__init__(
            self,
            command=ParallelCmds([merge, command.finalize()]),
            description="rescaling %s using mapDamage"
            % (describe_files(merge.input_files),),
            dependencies=dependencies,
        )

    def _setup(self, config, temp):
        CommandNode._setup(self, config, temp)
        for fname in ("Stats_out_MCMC_correct_prob.csv",):
            relpath = os.path.join(self._directory, fname)
            abspath = os.path.abspath(relpath)
            os.symlink(abspath, os.path.join(temp, fname))


# Minimal PDF written if Length_plot.pdf wasn't generated
_DUMMY_LENGTH_PLOT_PDF = """%PDF-1.4

1 0 obj
 <</Type /Font /Subtype /Type1 /Encoding /WinAnsiEncoding /BaseFont /Courier >>
endobj

2 0 obj
 <</Parent 4 0 R /MediaBox[0 0 450 50] /Type /Page /Contents[3 0 R ] /Resources 5 0 R >>
endobj

3 0 obj
 <</Length 138 >>
stream
  BT
    /F0 18 Tf
    20 10 Td
    (Input file(s) did not contain SE reads.) Tj
    0 20 Td
    (Length_plot.pdf not generated:) Tj
  ET
endstream
endobj

4 0 obj
 <</Type /Pages /Count 1 /Kids[2 0 R ]>>
endobj

5 0 obj
 <</ProcSet[/PDF /Text] /Font <</F0 1 0 R >>
>>
endobj

6 0 obj
 <</Type /Catalog /Pages 4 0 R >>
endobj

xref
0 7
0000000000 65535 f
0000000010 00000 n
0000000106 00000 n
0000000211 00000 n
0000000400 00000 n
0000000457 00000 n
0000000521 00000 n
trailer
 <</Size 7 /Root 6 0 R >>

startxref
571
%%EOF
"""
